from __future__ import annotations
from contextlib import contextmanager
from functools import wraps, cached_property
import weakref
import types
from typing import (
    Any,
    TYPE_CHECKING,
    Callable,
    Iterator,
    TypeVar,
    overload,
    Generic,
    Union,
)
from abc import ABCMeta
from typing_extensions import Literal, _AnnotatedAlias
from pathlib import Path
import datetime
import sys
from enum import Enum
from magicgui.widgets import create_widget, Container
from magicgui.widgets._bases import Widget, ValueWidget, ContainerWidget
from magicgui.widgets._bases.value_widget import UNSET
from psygnal import SignalInstance

from .utils import (
    argcount,
    is_instance_method,
    method_as_getter,
    Tooltips,
    eval_attribute,
)
from ._gui.mgui_ext import Action, WidgetAction

if TYPE_CHECKING:
    from magicgui.widgets._protocols import WidgetProtocol
    from typing_extensions import Self
    from ._gui._base import MagicTemplate
    from ._gui.mgui_ext import AbstractAction

    _M = TypeVar("_M", bound=MagicTemplate)

if sys.version_info >= (3, 10):
    from typing import _BaseGenericAlias
else:
    from typing_extensions import _BaseGenericAlias


class _FieldObject:
    name: str
    tooltip: str

    def get_widget(self, obj: Any) -> Widget:
        raise NotImplementedError()


_W = TypeVar("_W", bound=Widget)
_V = TypeVar("_V", bound=object)


class MagicField(_FieldObject, Generic[_W, _V]):
    """
    Field class for magicgui construction.

    This object is compatible with dataclass. MagicField object is in "ready for
    widget construction" state.
    """

    default_object = object()

    def __init__(
        self,
        value: Any = UNSET,
        name: str | None = None,
        label: str | None = None,
        annotation: Any = None,
        widget_type: type | str | None = None,
        options: dict[str, Any] | None = None,
        record: bool = True,
        constructor: Callable[..., Widget] | None = None,
    ):
        if options is None:
            options = {}
        if value is UNSET:
            value = options.pop("value", UNSET)

        if constructor is None:

            def _create_widget(obj):
                return create_widget(
                    self.value,
                    name=self.name,
                    label=self.label,
                    annotation=self.annotation,
                    widget_type=self.widget_type,
                    options=self.options,
                )

            constructor = _create_widget

        self.value = value
        self.name = name
        self.label = label
        self.annotation = annotation
        self.options = options
        self._widget_type = widget_type
        self._constructor = constructor
        self._callbacks: list[Callable] = []
        self._guis: dict[int, _M] = {}
        self._record = record

        # MagicField has to remenber the first class that referred to itself so
        # that it can "know" the namespace it belongs to.
        self._parent_class: type | None = None

    def __repr__(self):
        attrs = ["value", "name", "widget_type", "record", "options"]
        kw = ", ".join(f"{a}={getattr(self, a)!r}" for a in attrs)
        return f"{self.__class__.__name__}({kw})"

    def __set_name__(self, owner: type, name: str) -> None:
        self._parent_class = owner
        if self.name is None:
            self.name = name

    @property
    def constructor(self) -> Callable[..., Widget]:
        """Get widget constructor."""
        return self._constructor

    @property
    def callbacks(self) -> tuple[Callable, ...]:
        """Return callbacks in an immutable way."""
        return tuple(self._callbacks)

    @property
    def record(self) -> bool:
        return self._record

    def construct(self, obj) -> Widget:
        """Construct a widget."""
        constructor = self.constructor
        _arg_choices = self.options.get("choices", None)
        if isinstance(_arg_choices, str):
            _arg_choices = eval_attribute(type(obj), _arg_choices)

        if is_instance_method(_arg_choices):
            self.options["choices"] = method_as_getter(obj, _arg_choices)
        try:
            if _is_subclass(constructor, Widget):
                widget = constructor(**self.options)
            else:
                widget = constructor(obj)
                if not isinstance(widget, Widget):
                    raise TypeError(
                        f"{self.__class__.__name__} {self.name} created non-widget "
                        f"object {type(widget)}."
                    )

        finally:
            if _arg_choices is not None:
                self.options["choices"] = _arg_choices

        return widget

    def copy(self) -> Self[_W, _V]:
        """Copy object."""
        return self.__class__(
            value=self.value,
            name=self.name,
            label=self.label,
            annotation=self.annotation,
            widget_type=self.widget_type,
            options=self.options,
            record=self.record,
            constructor=self.constructor,
        )

    @classmethod
    def from_callable(cls, func: Callable[..., _W]) -> Self[_W, Any]:
        """Use a function as the constructor of MagicField."""
        return cls(constructor=func)

    def get_widget(self, obj: Any) -> _W:
        """
        Get a widget from ``obj``. This function will be called every time MagicField is referred
        by ``obj.field``.
        """
        from ._gui import MagicTemplate

        obj_id = id(obj)
        if obj_id in self._guis.keys():
            widget = self._guis[obj_id]
        else:
            widget = self.construct(obj)
            widget.name = self.name
            self._guis[obj_id] = widget

            if isinstance(widget, (ValueWidget, ContainerWidget)):
                if isinstance(obj, MagicTemplate):
                    _def = _define_callback_gui
                else:
                    _def = _define_callback
                for callback in self._callbacks:
                    # funcname = callback.__name__
                    widget.changed.connect(_def(obj, callback))

        return widget

    def get_action(self, obj: Any) -> AbstractAction:
        """
        Get an action from ``obj``. This function will be called every time MagicField is referred
        by ``obj.field``.
        """
        from ._gui import MagicTemplate

        obj_id = id(obj)
        if obj_id in self._guis.keys():
            action = self._guis[obj_id]
        else:
            if type(self.value) is bool or self.annotation is bool:
                # we should not use "isinstance" or "issubclass" because subclass
                # may be mapped to different widget by users.
                value = False if self.value is UNSET else self.value
                action = Action(
                    checkable=True,
                    checked=value,
                    text=self.name.replace("_", " "),
                    name=self.name,
                )
                for k, v in self.options.items():
                    setattr(action, k, v)
            else:
                widget = self.construct(obj)
                widget.name = self.name
                action = WidgetAction(widget)

            self._guis[obj_id] = action

            if action.support_value:
                if isinstance(obj, MagicTemplate):
                    _def = _define_callback_gui
                else:
                    _def = _define_callback
                for callback in self._callbacks:
                    # funcname = callback.__name__
                    action.changed.connect(_def(obj, callback))

        return action

    def as_getter(self, obj: Any) -> Callable[[Any], _V]:
        """Make a function that get the value of Widget or Action."""
        return lambda w: self._guis[id(obj)].value

    def as_remote_getter(self, obj: Any):
        """Called when a MagicField is used in Bound method."""
        qualname = self._parent_class.__qualname__
        _LOCALS = "<locals>."
        if _LOCALS in qualname:
            qualname = qualname.split(_LOCALS)[-1]
        clsnames = qualname.split(".")

        def _func(w):
            # First we have to know where (which instance) MagicField came from.
            if obj.__class__.__name__ not in clsnames:
                ns = ".".join(clsnames)
                raise ValueError(
                    f"Method {self.name!r} is in namespace {ns!r}, so it is invisible "
                    f"from magicclass {obj.__class__.__qualname__!r}."
                )
            i = clsnames.index(type(obj).__name__) + 1
            ins = obj
            for clsname in clsnames[i:]:
                ins = getattr(ins, clsname, ins)

            # Now, ins is an instance of parent class.
            # Extract correct widget from MagicField
            _field_widget = self.get_widget(ins)
            if not hasattr(_field_widget, "value"):
                raise TypeError(
                    f"MagicField {self.name} does not return ValueWidget "
                    "thus cannot be used as a bound value."
                )
            return self.as_getter(ins)(w)

        return _func

    @overload
    def __get__(self, obj: Literal[None], objtype=None) -> MagicField[_W, _V]:
        ...

    @overload
    def __get__(self, obj: Any, objtype=None) -> _W:
        ...

    def __get__(self, obj, objtype=None):
        """Get widget for the object."""
        if obj is None:
            return self
        return self.get_widget(obj)

    def __set__(self, obj, value) -> None:
        raise AttributeError(f"Cannot set value to {self.__class__.__name__}.")

    def ready(self) -> bool:
        """Return true if field is ready to create widgets."""
        return not self.not_ready()

    def not_ready(self) -> bool:
        return (
            self.value is UNSET
            and self.annotation is None
            and self.widget_type is None
            and "choices" not in self.options
        )

    def to_widget(self) -> _W:
        """
        Create a widget from the field.

        Returns
        -------
        Widget
            Widget object that is ready to be inserted into Container.

        Raises
        ------
        ValueError
            If there is not enough information to build a widget.
        """
        return self.get_widget(self.default_object)

    def to_action(self) -> Action | WidgetAction[_W]:
        """
        Create a menu action or a menu widget action from the field.

        Returns
        -------
        Action or WidgetAction
            Object that can be added to menu.

        Raises
        ------
        ValueError
            If there is not enough information to build an action.
        """
        return self.get_action(self.default_object)

    def connect(self, func: Callable) -> Callable:
        """Set callback function to "ready to connect" state."""
        if not callable(func):
            raise TypeError("Cannot connect non-callable object")
        self._callbacks.append(func)
        return func

    def disconnect(self, func: Callable) -> None:
        """
        Disconnect callback from the field.
        This method does NOT disconnect callbacks from widgets that are
        already created.
        """
        i = self._callbacks.index(func)
        self._callbacks.pop(i)
        return None

    def wraps(
        self,
        method: Callable | None = None,
        *,
        template: Callable | None = None,
        copy: bool = False,
    ):
        """
        Call the ``wraps`` class method of magic class.

        This method is needed when a child magic class is defined outside the main magic
        class, and integrated into the main magic class by ``field`` function, like below

        .. code-block:: python

            @magicclass
            class B:
                def func(self): ...  # pre-definition

            @magicclass
            class A:
                b = field(B)

                @b.wraps
                def func(self):
                    # do something

        Parameters
        ----------
        method : Callable, optional
            Method of parent class.
        template : Callable, optional
            Function template for signature.
        copy: bool, default is False
            If true, wrapped method is still enabled.

        Returns
        -------
        Callable
            Same method as input, but has updated signature.
        """
        from ._gui._base import BaseGui

        cls = self.constructor
        if not (isinstance(cls, type) and issubclass(cls, BaseGui)):
            raise TypeError(
                "The wraps method cannot be used for any objects but magic class."
            )
        return cls.wraps(method=method, template=template, copy=copy)

    @property
    def tooltip(self) -> str:
        """Get tooltip of returned widgets."""
        return self.options.get("tooltip", "")

    @tooltip.setter
    def tooltip(self, value):
        self.options.update(tooltip=value)

    @property
    def enabled(self) -> bool:
        """Get interactivity of returned widgets."""
        return self.options.get("enabled", True)

    @enabled.setter
    def enabled(self, value: bool):
        self.options.update(enabled=value)

    @property
    def visible(self) -> bool:
        """Get visibility of returned widgets."""
        return self.options.get("visible", True)

    @visible.setter
    def visible(self, value: bool):
        self.options.update(visible=value)

    @property
    def widget_type(self) -> type[Widget]:
        """Return type of the resulting widget."""
        return self._widget_type


class MagicValueField(MagicField[_W, _V]):
    """
    Field class for magicgui construction. Unlike MagicField, object of this class
    always returns value itself.
    """

    @overload
    def __get__(self, obj: Literal[None], objtype=None) -> MagicValueField[_W, _V] | _V:
        ...

    @overload
    def __get__(self, obj: Any, objtype=None) -> _V:
        ...

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return self._postgethook(obj, self.get_widget(obj).value)

    def __set__(self, obj: _M, value: _V) -> None:
        if obj is None:
            raise AttributeError(f"Cannot set {self.__class__.__name__}.")
        self.get_widget(obj).value = self._presethook(obj, value)

    def post_get_hook(self, hook: Callable[[Any, _V], Any] | Callable[[_V], Any]):
        """
        Define a post-get hook for the field.

        If a post-get hook is set, value will always be converted before returned.
        Following example shows how to convert ``x`` to a float every time it gets
        accessed.

        >>> class A:
        >>>     x = vfield(int)
        >>>     @x.post_get_hook
        >>>     def _x_get(self, value):
        >>>         return float(int)

        Parameters
        ----------
        hook : callable
            Post-get hook function.
        """
        if not callable(hook):
            raise TypeError("Post-get hook must be callable.")
        if is_instance_method(hook):
            self._postgethook = hook
        else:
            self._postgethook = lambda _, x: hook(x)
        return hook

    def pre_set_hook(self, hook: Callable[[Any, Any], _V] | Callable[[Any], _V]):
        """
        Define a pre-set hook for the field.

        If a pre-set hook is set, value will always be converted before being set
        to the widget value. Following example shows how to convert ``x`` to a
        string before setting the value

        >>> class A:
        >>>     x = vfield(str)
        >>>     @x.pre_set_hook
        >>>     def _x_set(self, value):
        >>>         return str(value)

        Parameters
        ----------
        hook : callable
            Pre-set hook function.
        """
        if not callable(hook):
            raise TypeError("Pre-set hook must be callable.")
        if is_instance_method(hook):
            self._presethook = hook
        else:
            self._presethook = lambda _, x: hook(x)
        return hook

    def _postgethook(self, obj, value):
        return value

    def _presethook(self, obj, value):
        return value


# magicgui symple types
_X = TypeVar(
    "_X",
    bound=Union[
        int,
        float,
        bool,
        str,
        Path,
        datetime.datetime,
        datetime.date,
        datetime.time,
        Enum,
        range,
        slice,
        list,
        tuple,
    ],
)


@overload
def field(
    obj: _X,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[ValueWidget, _X]:
    ...


@overload
def field(
    widget_type: type[_W],
    *,
    name: str | None = None,
    label: str | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[_W, Any]:
    ...


@overload
def field(
    obj: type[_X],
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[ValueWidget, _X]:
    ...


@overload
def field(
    gui_class: type[_M],
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[_M, Any]:
    ...


@overload
def field(
    obj: Any,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: type[_W] = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[_W, Any]:
    ...


@overload
def field(
    obj: Any,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[Widget, Any]:
    ...


def field(
    obj: Any = UNSET,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicField[Widget, Any]:
    """
    Make a MagicField object.

    >>> i = field(1)
    >>> i = field(widget_type="Slider")

    Parameters
    ----------
    obj : Any, default is UNSET
        Reference to determine what type of widget will be created. If Widget
        subclass is given, it will be used as is. If other type of class is given,
        it will used as type annotation. If an object (not type) is given, it will
        be assumed to be the default value.
    name : str, optional
        Name of the widget.
    label : str, optional
        Label of the widget.
    widget_type : str, optional
        Widget type. This argument will be sent to ``create_widget`` function.
    options : WidgetOptions, optional
        Widget options. This parameter will be passed to the ``options`` keyword
        argument of ``create_widget``.
    record : bool, default is True
        A magic-class specific parameter. If true, record value changes as macro.

    Returns
    -------
    MagicField
    """
    return _get_field(obj, name, label, widget_type, options, record, MagicField)


@overload
def vfield(
    obj: _X,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[ValueWidget, _X]:
    ...


@overload
def vfield(
    widget_type: type[_W],
    *,
    name: str | None = None,
    label: str | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[_W, Any]:
    ...


@overload
def vfield(
    annotation: type[_X],
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[ValueWidget, _X]:
    ...


@overload
def vfield(
    obj: Any,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: type[_W] = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[_W, Any]:
    ...


@overload
def vfield(
    obj: Any,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | type[Widget] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[Widget, Any]:
    ...


def vfield(
    obj: Any = UNSET,
    *,
    name: str | None = None,
    label: str | None = None,
    widget_type: str | type[WidgetProtocol] | None = None,
    options: dict[str, Any] = {},
    record: bool = True,
) -> MagicValueField[Widget, Any]:
    """
    Make a MagicValueField object.

    >>> i = vfield(1)
    >>> i = vfield(widget_type="Slider")

    Unlike MagicField, value itself can be accessed.

    >>> ui.i      # int is returned
    >>> ui.i = 3  # set value to the widget.

    Parameters
    ----------
    obj : Any, default is UNSET
        Reference to determine what type of widget will be created. If Widget
        subclass is given, it will be used as is. If other type of class is given,
        it will used as type annotation. If an object (not type) is given, it will
        be assumed to be the default value.
    name : str, optional
        Name of the widget.
    label : str, optional
        Label of the widget.
    widget_type : str, optional
        Widget type. This argument will be sent to ``create_widget`` function.
    options : WidgetOptions, optional
        Widget options. This parameter will be passed to the ``options`` keyword
        argument of ``create_widget``.
    record : bool, default is True
        A magic-class specific parameter. If true, record value changes as macro.

    Returns
    -------
    MagicValueField
    """
    return _get_field(obj, name, label, widget_type, options, record, MagicValueField)


def widget_property(func: Callable[..., _W]) -> MagicValueField[_W, Any]:
    """
    Create a widget property from a function.

    Generally, this decorator will be used in line with ``HasFields`` trait.
    If a widget has to be initialized depending on the state of the instance,
    you can define a field using a method.

    >>> class A(HasFields):
    >>>     def __init__(self, max: int = 10):
    >>>         self._max = max
    >>>     @widget_property
    >>>     def param(self):
    >>>         return SpinBox(value=1, max=self._max)

    In this example, ``param`` is a field object similar to those defined using
    ``vfield`` but has a variable argument ``max``.

    >>> a = A(max=20)
    >>> a.param  # Out: 1
    >>> a.param = 4  # OK
    >>> a.widgets.param  # SpinBox with value=1 and max=20

    Parameters
    ----------
    func : callable
        Widget constructor function.

    Returns
    -------
    MagicValueField
        A field with given constructor.
    """
    return MagicValueField.from_callable(func)


def _get_field(
    obj,
    name: str,
    label: str,
    widget_type: str | type[WidgetProtocol] | None,
    options: dict[str, Any],
    record: bool,
    field_class: type[MagicField],
) -> MagicField:
    if not isinstance(options, dict):
        raise TypeError(f"Field options must be a dict, got {type(options)}")
    options = options.copy()
    kwargs = dict(
        name=name, label=label, record=record, annotation=None, options=options
    )
    if isinstance(obj, (type, _BaseGenericAlias)):
        if isinstance(obj, _AnnotatedAlias):
            from magicgui.signature import split_annotated_type

            tp, widget_option = split_annotated_type(obj)
            kwargs.update(annotation=tp)
            options.update(**widget_option)
        if _is_subclass(obj, Widget):
            if widget_type is not None:
                raise ValueError("Cannot specify Widget type twice.")
            f = field_class(constructor=obj, widget_type=obj, **kwargs)
        else:
            if kwargs["annotation"] is None:
                kwargs.update(annotation=obj)
            f = field_class(widget_type=widget_type, **kwargs)
    elif obj is UNSET:
        f = field_class(widget_type=widget_type, **kwargs)
    else:
        f = field_class(value=obj, widget_type=widget_type, **kwargs)

    return f


def _is_subclass(obj: Any, class_or_tuple):
    try:
        return issubclass(obj, class_or_tuple)
    except Exception:
        return False


def _define_callback(self: Any, callback: Callable):
    """Define a callback function from a method."""
    return callback.__get__(self)


def _define_callback_gui(self: MagicTemplate, callback: Callable):
    """Define a callback function from a method of a magic-class."""

    *_, clsname, funcname = callback.__qualname__.split(".")
    mro = self.__class__.__mro__
    for base in mro:
        if base.__name__ == clsname:
            _func: Callable = getattr(base, funcname).__get__(self)
            _func = _normalize_argcount(_func)

            def _callback(v):
                with self.macro.blocked():
                    _func(v)
                return None

            break
    else:

        def _callback(v):
            # search for parent instances that have the same name.
            current_self = self
            while not (
                hasattr(current_self, funcname)
                and current_self.__class__.__qualname__.split(".")[-1] == clsname
            ):
                current_self = current_self.__magicclass_parent__
            _func = _normalize_argcount(getattr(current_self, funcname))

            with self.macro.blocked():
                _func(v)
            return None

    return _callback


def _normalize_argcount(func: Callable) -> Callable[[Any], Any]:
    if argcount(func) == 0:
        return lambda v: func()
    return func


class _FieldGroupMeta(ABCMeta):
    _fields: dict[str, MagicField]

    def __new__(
        fcls: type,
        name: str,
        bases: tuple,
        namespace: dict,
        **kwds,
    ) -> _FieldGroupMeta:
        cls: _FieldGroupMeta = type.__new__(fcls, name, bases, namespace, **kwds)
        _tooltips = Tooltips(cls)
        _fields: dict[str, MagicField] = {}
        for base in cls.__mro__[1:-1]:
            if type(base) is _FieldGroupMeta:
                _fields.update(base._fields)

        for k, v in namespace.items():
            if isinstance(v, _FieldObject):
                if k in _View._METHODS:
                    raise ValueError(f"Attribute {k} cannot be used in HasFields.")
                _fields[k] = v
                if not v.tooltip:
                    v.tooltip = _tooltips.attributes.get(k, "")

        cls._fields = types.MappingProxyType(_fields)

        return cls


class HasFields(metaclass=_FieldGroupMeta):
    """
    A trait implemented with widgets and signals.

    Subclasses can easily handle widgets and their value-change signals using
    the same attribute names.

    >>> class A(HasFields):
    >>>     a = vfield(int)
    >>>     b = vfield(str)
    >>> ins = A()
    >>> ins.a  # type: int
    >>> ins.widgets.a  # type: SpinBox
    >>> ins.signals.a  # type: SignalInstance

    """

    @cached_property
    def widgets(self) -> WidgetView:
        """Return a view of widgets."""
        return WidgetView(self)

    @cached_property
    def signals(self) -> SignalView:
        """Return a view of signals."""
        return SignalView(self)

    def __repr__(self) -> str:
        """List up child widgets."""
        _repr = ",\n\t".join(
            f"{name} = {repr(wdt)}" for name, wdt in self.widgets.iteritems()
        )
        return f"{self.__class__.__name__}(\n\t{_repr}\n)"


# NOTE: Typing of FieldGroup is tricky. When it is referred to via __get__, it
# must return a object of same type as itself to guarantee the child fields are
# also defined there. Thus, FieldGroup must inherit magicgui Container although
# the original container will never be used.
class FieldGroup(Container, HasFields, _FieldObject):
    def __init__(
        self,
        layout: str = "vertical",
        labels: bool = True,
        name: str | None = None,
        **kwargs,
    ):
        widgets = [fld.get_widget(self) for fld in self.__class__._fields.values()]
        super().__init__(
            layout=layout, widgets=widgets, labels=labels, name=name, **kwargs
        )
        self._containers: dict[int, Self] = {}
        self._callbacks: list[Callable] = []

    def __init_subclass__(cls) -> None:
        if "__init__" not in cls.__dict__.keys():
            return

        cls.__base_init__ = cls.__init__

        @wraps(cls.__init__)
        def __init__(self: cls, *args, **kwargs):
            self.__input_arguments = (args, kwargs)
            cls.__base_init__(self, *args, **kwargs)

        def __newlike__(self):
            args, kwargs = self.__input_arguments
            return cls(*args, **kwargs)

        cls.__init__ = __init__
        cls.__newlike__ = __newlike__

    def __newlike__(self) -> Self:
        """
        Make a copy of a FieldGroup.

        This method needs override if __init__ is overrided in a subclass.
        """
        return self.__class__(
            layout=self.layout,
            labels=self.labels,
            label=self.label,
            enabled=self.enabled,
            name=self.name,
            tooltip=self.tooltip,
        )

    def __set_name__(self, owner: type, name: str):
        """Set variable name as the container's name."""
        # self._parent_class = owner
        if self.name is None:
            self.name = name

    # Unlike Container, `self.x = value` should be allowed because `x` can be a value field.
    __setattr__ = object.__setattr__

    def copy(self) -> Self:
        """Copy widget."""
        wdt = self.__newlike__()
        for callback in self._callbacks:
            wdt.connect(callback)
        return wdt

    @property
    def callbacks(self) -> tuple[Callable, ...]:
        """Return callbacks in an immutable way."""
        return tuple(self._callbacks)

    @overload
    def __get__(self, obj: Literal[None], objtype: Any | None = None) -> Self:
        ...

    @overload
    def __get__(self, obj: Any, objtype: Any | None = None) -> Self:
        ...

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return self.get_widget(obj)

    def __set__(self, obj, value) -> None:
        raise AttributeError(f"Cannot set value to {self.__class__.__name__}.")

    def get_widget(self, obj) -> Container:
        _id = id(obj)
        wdt = self._containers.get(_id, None)
        if wdt is None:
            wdt = self.copy()
            self._containers[_id] = wdt
        return wdt

    def connect(self, callback: Callable):
        # self.changed.connect(callback)  NOTE: the original container doesn't need signals!
        self._callbacks.append(callback)
        return callback


class _View:
    _METHODS = (
        "iternames",
        "iterwidgets",
        "itersignals",
    )

    def __init__(self, obj: HasFields):
        self._obj_ref = weakref.ref(obj)

    def __repr__(self) -> str:
        _it = zip(self.iternames(), self.iterwidgets())
        _repr = ",\n\t".join(f"{name} = {repr(wdt)}" for name, wdt in _it)
        return f"{self.__class__.__name__}(\n\t{_repr}\n)"

    def iternames(self) -> Iterator[str]:
        """Iterate widget names."""
        return iter(self._obj_ref().__class__._fields.keys())

    def iterwidgets(self) -> Iterator[Widget]:
        """Iterate widgets."""
        obj = self._obj_ref()
        for fld in obj.__class__._fields.values():
            wdt = fld.get_widget(obj)
            yield wdt

    def itersignals(self, skip_undef: bool = False) -> Iterator[SignalInstance | None]:
        """Iterate value-changed signals."""
        obj = self._obj_ref()
        for fld in obj.__class__._fields.values():
            wdt = fld.get_widget(obj)
            sig = getattr(wdt, "changed", None)
            if isinstance(sig, SignalInstance):
                yield sig
            elif not skip_undef:
                yield None


class WidgetView(_View):
    """View of widgets."""

    def __getattr__(self, name: str) -> Widget:
        obj = self._obj_ref()
        fld = obj.__class__._fields.get(name, None)
        if isinstance(fld, _FieldObject):
            return fld.get_widget(obj)
        raise AttributeError(f"{obj!r} does not have attribute {name!r}.")

    def __getitem__(self, key: str | int) -> Widget:
        """Similar to Container's __getitem__."""
        if isinstance(key, int):
            obj = self._obj_ref()
            key = list(obj.__class__._fields.keys())[key]
        try:
            wdt = self.__getattr__(key)
        except AttributeError:
            raise KeyError(key)
        return wdt

    def iteritems(self) -> Iterator[tuple[str, Widget]]:
        """Iterate widget names and widgets themselves."""
        return iter(zip(self.iternames(), self.iterwidgets()))

    def __iter__(self) -> Iterator[Widget]:
        """Iterate widgets."""
        return self.iterwidgets()

    def as_container(
        self,
        layout: str = "vertical",
        labels: bool = True,
        keys: list[str] | None = None,
        **kwargs,
    ) -> Container:
        """Convert view into a Container widget."""
        if keys is None:
            widgets = list(self)
        else:
            widgets = [getattr(self, name) for name in keys]
        return Container(layout=layout, widgets=widgets, labels=labels, **kwargs)

    def emit_all(self) -> None:
        """Emit all the signals with current value."""
        for wdt, sig in zip(self.iterwidgets(), self.itersignals()):
            if sig is not None:
                sig.emit(wdt.value)


class SignalView(_View):
    """View of signals."""

    def __getattr__(self, name: str) -> SignalInstance:
        obj = self._obj_ref()
        fld = obj.__class__._fields.get(name, None)
        if isinstance(fld, _FieldObject):
            wdt = fld.get_widget(obj)
            sig = getattr(wdt, "changed", None)
            if not isinstance(sig, SignalInstance):
                raise AttributeError(f"Widget {wdt!r} does not have 'changed' signal")
            return sig
        raise AttributeError(f"{obj!r} does not have attribute {name!r}.")

    def __getitem__(self, key: str | int) -> SignalInstance:
        """Similar to list.__getitem__."""
        if isinstance(key, int):
            obj = self._obj_ref()
            key = list(obj.__class__._fields.keys())[key]
        try:
            wdt = self.__getattr__(key)
        except AttributeError:
            raise KeyError(key)
        return wdt

    def iteritems(self) -> Iterator[tuple[str, SignalInstance | None]]:
        """Iterate widget names and value-changed signals."""
        return iter(zip(self.iternames(), self.itersignals()))

    def __iter__(self) -> Iterator[SignalInstance | None]:
        return self.itersignals()

    def block(self) -> None:
        """Block all the signals."""
        for sig in self.itersignals(skip_undef=True):
            sig.block()

    def unblock(self) -> None:
        """Unblock all the signals."""
        for sig in self.itersignals(skip_undef=True):
            sig.unblock()

    @contextmanager
    def blocked(self):
        """Temporarly block all signals."""
        self.block()
        try:
            yield
        finally:
            self.unblock()



from __future__ import annotations
from functools import wraps
from typing import Any, Callable, TYPE_CHECKING
from dask.diagnostics import Callback as DaskCallback
from psygnal import Signal
from superqt.utils import FunctionWorker, GeneratorWorker, create_worker

from ...utils import move_to_screen_center, QtSignal
from ...utils.qthreading import (
    CallbackList,
    DefaultProgressBar,
    thread_worker,
    ProgressDict,
)

if TYPE_CHECKING:
    from ..._gui import BaseGui


class DaskProgressBar(DefaultProgressBar, DaskCallback):
    """A progress bar widget for dask computation."""

    computed = Signal(object)

    def __init__(
        self,
        max: int = 100,
        minimum: float = 0.5,
        dt: float = 0.1,
    ):
        self._minimum = minimum
        self._dt = dt
        self._frac = 0.0
        self._n_computation = 0
        super().__init__(max=max)
        self.footer[1].visible = self.footer[2].visible = False
        self._computed_signal = QtSignal()
        self._computed_signal.connect(self._on_computed)

    def __enter__(self):
        self._n_computation = 0
        self._on_timer_updated()
        return super().__enter__()

    def _start(self, dsk):
        self._state = None
        self._frac = 0.0
        self._n_computation += 1
        self._start_thread()
        return None

    def _on_computed(self, result):
        s = self._state
        if not s:
            self._frac = 0.0
        else:
            ndone = len(s["finished"])
            ntasks = sum(len(s[k]) for k in ["ready", "waiting", "running"]) + ndone
            if ndone <= ntasks:
                self._frac = ndone / ntasks if ntasks else 0.0
        self.pbar.value = self.max * self._frac
        self.computed.emit(result)
        return None

    def _pretask(self, key, dsk, state):
        self._state = state
        return None

    def _posttask(self, key, result, dsk, state, worker_id):
        self._computed_signal.emit(result)
        self._time_signal.emit()
        return None

    def _finish(self, dsk, state, errored):
        self._frac = 1.0
        self._running = False
        self._thread_timer.join()
        self._timer.reset()
        return None

    def _on_timer_updated(self, _=None):
        if self._n_computation > 1:
            _prefix = f"({self._n_computation}) "
        else:
            _prefix = ""
        if self._timer.sec < 3600:
            self.time_label.value = _prefix + self._timer.format_time(
                "{min:0>2}:{sec:0>2}"
            )
        else:
            self.time_label.value = _prefix + self._timer.format_time()
        return None

    def set_worker(self, worker: GeneratorWorker | FunctionWorker):
        """Set currently running worker."""
        self._worker = worker
        if isinstance(self._worker, GeneratorWorker):
            raise TypeError("Cannot set generator.")
        self.footer[1].visible = False
        self.footer[2].visible = False
        self._time_signal.emit()
        return None


class dask_thread_worker(thread_worker):
    """
    Create a dask's worker in a superqt/napari style.

    This thread worker class can monitor the progress of dask computation.
    Callback function connected to ``computed`` signal will get called when any one
    of the tasks are finished. The returned value of the task will be sent to the
    callback argument. The returned value is useful if delayed functions are computed
    but it is not always meaningful when dask mapping functions such as ``map_blocks``
    is used. Unlike standard ``thread_worker``, you should not specify ``total``
    parameter since dask progress bar knows it.

    Examples
    --------

    .. code-block:: python

        @magicclass
        class A:
            @dask_thread_worker
            def func(self):
                arr = da.random.random((30000, 30000))
                da.mean(arr).compute()

            @func.computed.connect
            def _on_computed(self, _=None):
                print("computed")

    """

    _DEFAULT_PROGRESS_BAR = DaskProgressBar
    _DEFAULT_TOTAL = 100

    def __init__(
        self,
        f: Callable | None = None,
        *,
        ignore_errors: bool = False,
        progress: ProgressDict | bool | None = True,
    ) -> None:
        super().__init__(f, ignore_errors=ignore_errors, progress=progress)
        self._callback_dict_["computed"] = CallbackList()

    @property
    def computed(self) -> CallbackList[Any]:
        return self._callback_dict_["computed"]

    def _create_method(self, gui: BaseGui):
        if self._progress is None:
            self._progress = {
                "pbar": None,
                "desc": "Progress",
                "total": 100,
            }
        else:
            self._progress["pbar"] = None

        return super()._create_method(gui)

    def _create_qt_worker(
        self, gui, *args, **kwargs
    ) -> FunctionWorker | GeneratorWorker:
        gui_id = id(gui)
        if gui_id in self._progressbars:
            pbar = self._progressbars[gui_id]
        else:
            pbar = self._DEFAULT_PROGRESS_BAR(max=self._DEFAULT_TOTAL)
            self._progressbars[gui_id] = pbar
            for c in self.computed._iter_as_method(gui):
                pbar.computed.connect(c)
            pbar.native.setParent(gui.native, self.__class__._WINDOW_FLAG)
            move_to_screen_center(pbar.native)

        worker = create_worker(
            self._define_function(pbar).__get__(gui),
            _ignore_errors=self._ignore_errors,
            _start_thread=False,
            *args,
            **kwargs,
        )

        return worker

    def _define_function(self, pbar):
        @wraps(self._func)
        def _wrapped(*args, **kwargs):
            with pbar:
                out = self._func(*args, **kwargs)
            return out

        return _wrapped



from magicclass import magicclass, magicmenu, magictoolbar
from magicclass.types import WidgetType

def _make_class(t: WidgetType):
    @magicclass(widget_type=t)
    class A:
        @magicmenu
        class Menu:
            pass

        @magictoolbar
        class Tool1:
            pass

        @magictoolbar
        class Tool2:
            pass

        @magicclass
        class B:
            def b1(self, a: int): ...
            def b2(self, a: int): ...

        def a1(self, a: int): ...

        @B.wraps
        def b1(self, a: int): ...

    return A


def test_all_works():
    for wtype in WidgetType._member_names_:
        ui = _make_class(wtype)()
        ui.show(run=False)
        ui[0]
        ui[1].changed()
        ui.B[0].changed()
        ui.B[1].changed()
        if hasattr(ui, "current_index"):
            ui.current_index = 1
        ui.close()



from __future__ import annotations
import numpy as np


def convert_color_code(c):
    if not isinstance(c, str):
        c = np.asarray(c) * 255
    return c


def to_rgba(pen) -> np.ndarray:
    rgba = pen.color().getRgb()
    return np.array(rgba) / 255



from __future__ import annotations
from typing import Callable, Iterable, Any, Generic, TypeVar
from qtpy.QtWidgets import (
    QPushButton,
    QAction,
    QWidgetAction,
    QToolButton,
    QWidget,
    QMenu,
)
from psygnal import Signal
from magicgui.widgets import PushButton
from magicgui.widgets._concrete import _LabeledWidget
from magicgui.widgets._bases import Widget, ValueWidget
from magicgui.backends._qtpy.widgets import QBaseButtonWidget
from ._function_gui import FunctionGuiPlus
from ._icon import get_icon

# magicgui widgets that need to be extended to fit into magicclass


class PushButtonPlus(PushButton):
    """A Qt specific PushButton widget with a magicgui bound."""

    def __init__(self, text: str | None = None, **kwargs):
        super().__init__(text=text, **kwargs)
        self.native: QPushButton
        self._icon = None
        self.mgui: FunctionGuiPlus | None = None  # tagged function GUI
        self._doc = ""
        self._unwrapped = False

    @property
    def running(self) -> bool:
        """Return true if embedded magicgui widget is running from GUI."""
        return getattr(self.mgui, "running", False)

    def set_shortcut(self, key):
        """Set keyboard shortcut to the button."""
        self.native.setShortcut(key)

    def reset_choices(self, *_: Any):
        """Reset child Categorical widgets."""
        if self.mgui is not None:
            self.mgui.reset_choices()

    @property
    def background_color(self):
        """The background color of the button."""
        return self.native.palette().button().color().getRgb()

    @background_color.setter
    def background_color(self, color: str | Iterable[float]):
        stylesheet = self.native.styleSheet()
        d = _stylesheet_to_dict(stylesheet)
        d.update({"background-color": _to_rgb(color)})
        stylesheet = _dict_to_stylesheet(d)
        self.native.setStyleSheet(stylesheet)

    @property
    def icon(self):
        """Get icon object."""
        return self._icon

    @icon.setter
    def icon(self, val):
        icon = get_icon(val)
        icon.install(self)
        self._icon = icon
        self.native.setIconSize(self.native.size())

    icon_path = icon

    @property
    def font_size(self):
        """Font size of the button."""
        return self.native.font().pointSize()

    @font_size.setter
    def font_size(self, size: int):
        font = self.native.font()
        font.setPointSize(size)
        self.native.setFont(font)

    @property
    def font_color(self):
        """Font color of the button."""
        return self.native.palette().text().color().getRgb()

    @font_color.setter
    def font_color(self, color: str | Iterable[float]):
        stylesheet = self.native.styleSheet()
        d = _stylesheet_to_dict(stylesheet)
        d.update({"color": _to_rgb(color)})
        stylesheet = _dict_to_stylesheet(d)
        self.native.setStyleSheet(stylesheet)

    @property
    def font_family(self):
        """Font family of the button."""
        return self.native.font().family()

    @font_family.setter
    def font_family(self, family: str):
        font = self.native.font()
        font.setFamily(family)
        self.native.setFont(font)

    def from_options(self, options: dict[str, Any] | Callable):
        """Set button design options using a dictionary or a function."""

        if callable(options):
            try:
                options = options.__signature__.caller_options
            except AttributeError:
                return None

        for k, v in options.items():
            setattr(self, k, v)
        return None


class _QToolButton(QBaseButtonWidget):
    def __init__(self):
        super().__init__(QToolButton)


class ToolButtonPlus(PushButtonPlus):
    """Buttons for toolbar in magic-class."""

    def __init__(self, text: str | None = None, **kwargs):
        kwargs["widget_type"] = _QToolButton
        ValueWidget.__init__(self, **kwargs)
        self.text = text or self.name.replace("_", " ")
        self.native: QToolButton

    def set_menu(self, qmenu: QMenu):
        """Set menu-like behavior to the tool button."""
        self.native.setMenu(qmenu)
        self.native.setPopupMode(QToolButton.InstantPopup)
        self.native.setIcon(qmenu.icon())  # icon have to be copied.

    def set_shortcut(self, key):
        """Set keyboard shortcut to the tool button."""
        self.native.setShortcut(key)

    def reset_choices(self, *_: Any):
        """Reset child Categorical widgets."""
        if self.mgui is not None:
            self.mgui.reset_choices()


class mguiLike:
    """Abstract class that provide magicgui.widgets like properties."""

    native: QWidget | QAction

    @property
    def parent(self):
        return self.native.parent()

    @parent.setter
    def parent(self, obj: mguiLike | Widget):
        self.native.setParent(obj.native)

    @property
    def name(self) -> str:
        return self.native.objectName()

    @name.setter
    def name(self, value: str):
        self.native.setObjectName(value)

    @property
    def tooltip(self) -> str:
        return self.native.toolTip()

    @tooltip.setter
    def tooltip(self, value: str):
        self.native.setToolTip(value)

    @property
    def enabled(self) -> bool:
        return self.native.isEnabled()

    @enabled.setter
    def enabled(self, value: bool):
        self.native.setEnabled(value)

    @property
    def visible(self) -> bool:
        return self.native.isVisible()

    @visible.setter
    def visible(self, value: bool):
        self.native.setVisible(value)

    @property
    def widget_type(self):
        return self.__class__.__name__


class AbstractAction(mguiLike):
    """
    QAction encapsulated class with a similar API as magicgui Widget.
    This class makes it easier to combine QMenu to magicgui.
    """

    changed = Signal(object)
    support_value: bool
    native: QAction | QWidgetAction

    @property
    def value(self):
        raise NotImplementedError()

    def from_options(self, options):
        raise NotImplementedError()


class Action(AbstractAction):
    support_value = True

    def __init__(
        self, *args, name: str = None, text: str = None, gui_only: bool = True, **kwargs
    ):
        self.native = QAction(*args, **kwargs)
        self.mgui: FunctionGuiPlus = None
        self._doc = ""
        self._unwrapped = False

        self._icon = None
        if text:
            self.text = text
        if name:
            self.native.setObjectName(name)
        self._callbacks = []

        self.native.triggered.connect(lambda: self.changed.emit(self.value))

    @property
    def running(self) -> bool:
        return getattr(self.mgui, "running", False)

    def set_shortcut(self, key):
        self.native.setShortcut(key)

    def reset_choices(self, *_: Any):
        """Reset child Categorical widgets."""
        if self.mgui is not None:
            self.mgui.reset_choices()

    @property
    def text(self) -> str:
        return self.native.text()

    @text.setter
    def text(self, value: str):
        self.native.setText(value)

    @property
    def value(self):
        return self.native.isChecked()

    @value.setter
    def value(self, checked: bool):
        self.native.setChecked(checked)

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, val):
        icon = get_icon(val)
        icon.install(self)
        self._icon = icon

    icon_path = icon

    def from_options(self, options: dict[str] | Callable):
        if callable(options):
            try:
                options = options.__signature__.caller_options
            except AttributeError:
                return None

        for k, v in options.items():
            setattr(self, k, v)
        return None


_W = TypeVar("_W", bound=Widget)


class WidgetAction(AbstractAction, Generic[_W]):
    """An Action class that contains a Widget object."""

    def __init__(self, widget: _W, label: str = None, parent=None):
        if not isinstance(widget, (Widget, mguiLike)):
            raise TypeError(
                f"The first argument must be a magicgui-like widget, got {type(widget)}"
            )

        self.native = QWidgetAction(parent)
        self.widget = widget
        name = widget.name
        self.native.setObjectName(name)
        self.label = label or name.replace("_", " ")
        self.text = getattr(widget, "text", None) or name.replace("_", " ")

        self.native.setDefaultWidget(widget.native)
        self.support_value = isinstance(
            widget, (ValueWidget, _LabeledWidget)
        ) or hasattr(widget, "value")

        if self.support_value:
            self.widget.changed.connect(lambda: self.changed.emit(self.value))

    @property
    def widget_type(self):
        return self.widget.widget_type

    @property
    def value(self):
        if self.support_value:
            return self.widget.value
        else:
            msg = (
                f"WidgetAction {self.name} has {type(self.widget)} as the default "
                "widget, which does not have value property."
            )
            raise AttributeError(msg)

    @value.setter
    def value(self, v):
        if self.support_value:
            self.widget.value = v
        else:
            msg = (
                f"WidgetAction {self.name} has {type(self.widget)} as the default "
                "widget, which does not have value property."
            )
            raise AttributeError(msg)

    @property
    def text(self) -> str:
        return self.native.text()

    @text.setter
    def text(self, value: str):
        self.native.setText(value)
        if hasattr(self.widget, "text"):
            self.widget.text = value

    def _labeled_widget(self):
        return self.widget._labeled_widget()

    def render(self):
        return self.widget.render()

    def _repr_png_(self):
        return self.widget._repr_png_()


class _LabeledWidgetAction(WidgetAction):
    widget: _LabeledWidget

    def __init__(self, widget: Widget, label: str = None):
        if not isinstance(widget, Widget):
            raise TypeError(f"The first argument must be a Widget, got {type(widget)}")

        _labeled_widget = _LabeledWidget(widget, label)
        super().__init__(_labeled_widget)
        self.name = widget.name

        # Strangely, visible.setter does not work for sliders.
        widget.native.setVisible(True)

    @classmethod
    def from_action(cls, action: WidgetAction):
        """Construct a labeled action using another action."""
        self = cls(action.widget, action.label)
        action.parent = self
        return self

    @property
    def label_width(self):
        return self.widget._label_widget.width

    @label_width.setter
    def label_width(self, width):
        self.widget._label_widget.min_width = width


def _to_rgb(color):
    if isinstance(color, str):
        from matplotlib.colors import to_rgb

        color = to_rgb(color)
    rgb = ",".join(str(max(min(int(c * 255), 255), 0)) for c in color)
    return f"rgb({rgb})"


def _stylesheet_to_dict(stylesheet: str):
    if stylesheet == "":
        return {}
    lines = stylesheet.split(";")
    d = dict()
    for line in lines:
        k, v = line.split(":")
        d[k.strip()] = v.strip()
    return d


def _dict_to_stylesheet(d: dict):
    stylesheet = [f"{k}: {v}" for k, v in d.items()]
    return ";".join(stylesheet)



from magicclass import magicclass, magicmenu, magictoolbar, field, MagicTemplate

def test_find_ancestor():
    @magicclass
    class B(MagicTemplate):
        def f(self):
            parent = self.find_ancestor(A)
            parent.i.value = 10

    @magicclass
    class A(MagicTemplate):
        i = field(int)
        b = field(B)

    ui = A()
    assert ui.i.value == 0
    ui.b.f()
    assert ui.i.value == 10

def test_add_dock_widget():
    @magicclass(widget_type="mainwindow")
    class Main(MagicTemplate):
        pass

    @magicclass
    class A(MagicTemplate):
        i = field(int)

    ui = Main()
    ui.add_dock_widget(A())
    ui.add_dock_widget(A())
    ui.add_dock_widget(A())


def test_post_append():
    from types import MethodType

    @magicclass
    class A:
        a = 0
        @magicmenu
        class Menu:
            pass
        @magictoolbar
        class Tool:
            pass

        def f(self):
            self.a = 1

    ui = A()

    @ui.append
    def g(): pass
    @ui.Menu.append
    def g(): pass
    @ui.Tool.append
    def g(): pass

    assert isinstance(ui.g, MethodType)
    ui["g"]
    assert isinstance(ui.Menu.g, MethodType)
    ui.Menu["g"]
    assert isinstance(ui.Tool.g, MethodType)
    ui.Tool["g"]

    @ui.append
    def f(): pass

    # the original method should not be updated
    ui.f()
    assert ui.a == 1



from __future__ import annotations

from typing import Iterable, Callable, TYPE_CHECKING
from magicgui import register_type, application as app
from magicgui.widgets import Container, ComboBox, Label, Widget
from magicgui.widgets._bases import CategoricalWidget

from napari.utils._magicgui import find_viewer_ancestor

if TYPE_CHECKING:
    from .types import Features, FeatureColumn


def get_features(widget: CategoricalWidget) -> list[tuple[str, Features]]:
    """Get all the non-empty feature data from the viewer."""
    viewer = find_viewer_ancestor(widget)
    if viewer is None:
        return []
    features: list[Features] = []
    for layer in viewer.layers:
        if len(feat := getattr(layer, "features", [])) > 0:
            features.append((layer.name, feat))
    return features


# Widget
class ColumnChoice(Container):
    def __init__(
        self,
        data_choices: Iterable[Features] | Callable[[Widget], Iterable[Features]],
        value=None,
        **kwargs,
    ):
        self._dataframe_cbox = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_cbox = ComboBox(choices=self._get_available_columns)
        _measure = app.use_app().get_obj("get_text_width")
        _label_l = Label(value='.features["')
        _label_l.max_width = _measure(_label_l.value)
        _label_r = Label(value='"]')
        _label_r.max_width = _measure(_label_r.value)

        super().__init__(
            layout="horizontal",
            widgets=[self._dataframe_cbox, _label_l, self._column_cbox, _label_r],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_cbox.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: Features = self._dataframe_cbox.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        self._column_cbox.choices = cols
        return None

    @property
    def value(self) -> FeatureColumn:
        df = self._dataframe_cbox.value
        return df[self._column_cbox.value]


class ColumnNameChoice(Container):
    """
    A container widget with a DataFrame selection and multiple column name selections.

    This widget is composed of two or more ComboBox widgets. The top one is to choose a
    DataFrame and the rest are to choose column names from the DataFrame. When the DataFrame
    selection changed, the column name selections will also changed accordingly.
    """

    def __init__(
        self,
        data_choices: Iterable[Features] | Callable[[Widget], Iterable[Features]],
        column_choice_names: Iterable[str],
        value=None,
        **kwargs,
    ):
        self._dataframe_cbox = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_names_cbox: list[ComboBox] = []
        for cn in column_choice_names:
            self._column_names_cbox.append(
                ComboBox(choices=self._get_available_columns, name=cn, nullable=True)
            )
        self._child_container = Container(
            widgets=self._column_names_cbox, layout="vertical"
        )
        self._child_container.margins = (0, 0, 0, 0)
        super().__init__(
            layout="vertical",
            widgets=[self._dataframe_cbox, self._child_container],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_cbox.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: Features = self._dataframe_cbox.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        for cbox in self._column_names_cbox:
            cbox.choices = cols
        return None

    @property
    def value(self) -> tuple[Features, list[str]]:
        df = self._dataframe_cbox.value
        colnames = [cbox.value for cbox in self._column_names_cbox]
        return (df, colnames)


def _register_mgui_types():
    from .types import Features, FeatureColumn, FeatureInfoInstance

    register_type(Features, choices=get_features, nullable=False)

    register_type(
        FeatureColumn,
        widget_type=ColumnChoice,
        data_choices=get_features,
        nullable=False,
    )

    register_type(
        FeatureInfoInstance,
        widget_type=ColumnNameChoice,
        data_choices=get_features,
    )



from __future__ import annotations
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Callable,
    TypeVar,
    overload,
)
from typing_extensions import ParamSpec

from .utils import FreeWidget

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from matplotlib.axes import Axes
    from matplotlib.lines import Line2D
    from matplotlib.collections import PathCollection
    from matplotlib.text import Text
    from matplotlib.quiver import Quiver
    from matplotlib.legend import Legend
    from numpy.typing import ArrayLike

    _P = ParamSpec("_P")
    _R = TypeVar("_R")

    def _inject_mpl_docs(f: Callable[_P, _R]) -> Callable[_P, _R]:
        plt_func = getattr(plt, f.__name__)
        plt_doc = getattr(plt_func, "__doc__", "")
        if plt_doc:
            f.__doc__ = (
                f"Copy of ``plt.{f.__name__}()``. Original docstring "
                f"is ...\n\n{plt_doc}"
            )
        return f

    import seaborn as sns
    from seaborn.axisgrid import Grid

    def _inject_sns_docs(f: Callable[_P, _R]) -> Callable[_P, _R]:
        sns_func = getattr(sns, f.__name__)
        sns_doc = getattr(sns_func, "__doc__", "")
        if sns_doc:
            f.__doc__ = (
                f"Copy of ``sns.{f.__name__}()``. Original docstring "
                f"is ...\n\n{sns_doc}"
            )
        return f

else:

    def _inject_mpl_docs(f: Callable[_P, _R]) -> Callable[_P, _R]:
        return f

    def _inject_sns_docs(f: Callable[_P, _R]) -> Callable[_P, _R]:
        return f


class Figure(FreeWidget):
    """A matplotlib figure canvas."""

    _docstring_initialized = False

    @overload
    def __init__(
        self,
        nrows: int = 1,
        ncols: int = 1,
        figsize: tuple[float, float] = (4.0, 3.0),
        style=None,
        **kwargs,
    ):
        ...

    @overload
    def __init__(
        self,
        fig: Figure,
        **kwargs,
    ):
        ...

    def __init__(
        self,
        nrows=1,
        ncols=1,
        figsize=(4.0, 3.0),
        style=None,
        **kwargs,
    ):

        if isinstance(nrows, int):
            import matplotlib as mpl
            import matplotlib.pyplot as plt

            backend = mpl.get_backend()
            try:
                mpl.use("Agg")
                if style is None:
                    fig, _ = plt.subplots(nrows, ncols, figsize=figsize)
                else:
                    with plt.style.context(style):
                        fig, _ = plt.subplots(nrows, ncols, figsize=figsize)
            finally:
                mpl.use(backend)
        else:
            fig = nrows

        from ._mpl_canvas import InteractiveFigureCanvas

        canvas = InteractiveFigureCanvas(fig)
        self.canvas = canvas
        super().__init__(**kwargs)
        self.set_widget(canvas)
        self.figure = fig
        self.min_height = 40
        self.min_width = 40
        self._inject_docs()

    def _reset_canvas(self, fig: Figure, draw: bool = True):
        """Create an interactive figure canvas and add it to the widget."""
        from ._mpl_canvas import InteractiveFigureCanvas

        canvas = InteractiveFigureCanvas(fig)
        if self.central_widget is not None:
            self.remove_widget(self.canvas)
        self.canvas = canvas
        self.set_widget(canvas)
        if draw:
            self.draw()

    def _inject_docs(self):
        if Figure._docstring_initialized:
            return

        import matplotlib.pyplot as plt

        for k, f in Figure.__dict__.items():
            if k.startswith("_") or isinstance(f, property):
                continue
            plt_func = getattr(plt, k, None)
            if plt_func is None:
                continue
            plt_doc = getattr(plt_func, "__doc__", "")
            if plt_doc:
                f.__doc__ = (
                    f"Copy of ``plt.{k}()``. Original docstring " f"is ...\n\n{plt_doc}"
                )

        Figure._docstring_initialized = True

    @_inject_mpl_docs
    def draw(self):
        self.figure.tight_layout()
        self.canvas.draw()

    @property
    def enabled(self) -> bool:
        """toggle interactivity of the figure canvas."""
        return self.canvas._interactive

    @enabled.setter
    def enabled(self, v: bool):
        self.canvas._interactive = bool(v)

    @property
    def mouse_click_callbacks(self) -> list[Callable]:
        return self.canvas._mouse_click_callbacks

    interactive = enabled  # alias

    @_inject_mpl_docs
    def clf(self) -> None:
        self.figure.clf()
        self.draw()

    @_inject_mpl_docs
    def cla(self) -> None:
        self.ax.cla()
        self.draw()

    @property
    def axes(self) -> list[Axes]:
        """List of matplotlib axes."""
        return self.figure.axes

    @property
    def ax(self) -> Axes:
        """The first matplotlib axis."""
        try:
            _ax = self.axes[0]
        except IndexError:
            _ax = self.figure.add_subplot(111)
        return _ax

    @_inject_mpl_docs
    def subplots(
        self, *args, **kwargs
    ) -> tuple[Figure, Axes] | tuple[Figure, list[Axes]]:
        self.clf()
        fig = self.figure
        axs = fig.subplots(*args, **kwargs)
        self.draw()
        return fig, axs

    @_inject_mpl_docs
    def savefig(self, *args, **kwargs) -> None:
        return self.figure.savefig(*args, **kwargs)

    @_inject_mpl_docs
    def plot(self, *args, **kwargs) -> list[Line2D]:
        lines = self.ax.plot(*args, **kwargs)
        self.draw()
        return lines

    @_inject_mpl_docs
    def scatter(self, *args, **kwargs) -> PathCollection:
        paths = self.ax.scatter(*args, **kwargs)
        self.draw()
        return paths

    @_inject_mpl_docs
    def hist(
        self, *args, **kwargs
    ) -> tuple[ArrayLike | list[ArrayLike], ArrayLike, list | list[list]]:
        out = self.ax.hist(*args, **kwargs)
        self.draw()
        return out

    @_inject_mpl_docs
    def text(self, *args, **kwargs) -> Text:
        text = self.ax.text(*args, **kwargs)
        self.draw()
        return text

    @_inject_mpl_docs
    def quiver(self, *args, data=None, **kwargs) -> Quiver:
        quiver = self.ax.quiver(*args, data=data, **kwargs)
        self.draw()
        return quiver

    @_inject_mpl_docs
    def axline(self, xy1, xy2=None, *, slope=None, **kwargs) -> Line2D:
        lines = self.ax.axline(xy1, xy2=xy2, slope=slope, **kwargs)
        self.draw()
        return lines

    @_inject_mpl_docs
    def axhline(self, y=0, xmin=0, xmax=1, **kwargs) -> Line2D:
        lines = self.ax.axhline(y, xmin, xmax, **kwargs)
        self.draw()
        return lines

    @_inject_mpl_docs
    def axvline(self, x=0, ymin=0, ymax=1, **kwargs) -> Line2D:
        lines = self.ax.axvline(x, ymin, ymax, **kwargs)
        self.draw()
        return lines

    @_inject_mpl_docs
    def xlim(self, *args, **kwargs) -> tuple[float, float]:
        ax = self.ax
        if not args and not kwargs:
            return ax.get_xlim()
        ret = ax.set_xlim(*args, **kwargs)
        self.draw()
        return ret

    @_inject_mpl_docs
    def ylim(self, *args, **kwargs) -> tuple[float, float]:
        ax = self.ax
        if not args and not kwargs:
            return ax.get_ylim()
        ret = ax.set_ylim(*args, **kwargs)
        self.draw()
        return ret

    @_inject_mpl_docs
    def imshow(self, *args, **kwargs) -> Axes:
        self.ax.imshow(*args, **kwargs)
        self.draw()
        return self.ax

    @_inject_mpl_docs
    def legend(self, *args, **kwargs) -> Legend:
        leg = self.ax.legend(*args, **kwargs)
        self.draw()
        return leg

    @_inject_mpl_docs
    def title(self, *args, **kwargs) -> Text:
        title = self.ax.set_title(*args, **kwargs)
        self.draw()
        return title

    @_inject_mpl_docs
    def xlabel(self, *args, **kwargs) -> None:
        self.ax.set_xlabel(*args, **kwargs)
        self.draw()
        return None

    @_inject_mpl_docs
    def ylabel(self, *args, **kwargs) -> None:
        self.ax.set_ylabel(*args, **kwargs)
        self.draw()
        return None

    @_inject_mpl_docs
    def xticks(self, ticks=None, labels=None, **kwargs) -> tuple[ArrayLike, list[Text]]:
        if ticks is None:
            locs = self.ax.get_xticks()
            if labels is not None:
                raise TypeError(
                    "xticks(): Parameter 'labels' can't be set "
                    "without setting 'ticks'"
                )
        else:
            locs = self.ax.set_xticks(ticks)

        if labels is None:
            labels = self.ax.get_xticklabels()
        else:
            labels = self.ax.set_xticklabels(labels, **kwargs)
        for l in labels:
            l.update(kwargs)
        self.draw()
        return locs, labels

    @_inject_mpl_docs
    def yticks(self, ticks=None, labels=None, **kwargs) -> tuple[ArrayLike, list[Text]]:
        if ticks is None:
            locs = self.ax.get_yticks()
            if labels is not None:
                raise TypeError(
                    "xticks(): Parameter 'labels' can't be set "
                    "without setting 'ticks'"
                )
        else:
            locs = self.ax.set_yticks(ticks)

        if labels is None:
            labels = self.ax.get_yticklabels()
        else:
            labels = self.ax.set_yticklabels(labels, **kwargs)
        for l in labels:
            l.update(kwargs)
        self.draw()
        return locs, labels

    @_inject_mpl_docs
    def twinx(self) -> Axes:
        return self.ax.twinx()

    @_inject_mpl_docs
    def twiny(self) -> Axes:
        return self.ax.twiny()

    @_inject_mpl_docs
    def box(self, on=None) -> None:
        if on is None:
            on = not self.ax.get_frame_on()
        self.ax.set_frame_on(on)
        return None

    @_inject_mpl_docs
    def xscale(self, scale=None) -> None:
        self.ax.set_xscale(scale)
        self.draw()
        return None

    @_inject_mpl_docs
    def yscale(self, scale=None) -> None:
        self.ax.set_yscale(scale)
        self.draw()
        return None

    @_inject_mpl_docs
    def autoscale(self, enable=True, axis="both", tight=None) -> None:
        self.ax.autoscale(enable=enable, axis=axis, tight=tight)
        self.draw()
        return None


def _use_seaborn_grid(f):
    """
    Some seaborn plot functions will create a new figure.
    This decorator provides a common way to update figure canvas in the widget.
    """

    @wraps(f)
    def func(self: SeabornFigure, *args, **kwargs):
        import matplotlib as mpl

        backend = mpl.get_backend()
        try:
            mpl.use("Agg")
            grid: Grid = f(self, *args, **kwargs)
        finally:
            mpl.use(backend)

        self._reset_canvas(grid.figure)
        return grid

    return func


class SeabornFigure(Figure):
    """
    A matplotlib figure canvas implemented with seaborn plot functions.

    Not all the seaborn plot functions are supported since some of them are
    figure-level functions and incompatible with specifying axes.
    """

    _docstring_initialized = False

    def __init__(
        self,
        nrows: int = 1,
        ncols: int = 1,
        figsize: tuple[float, float] = (4.0, 3.0),
        style=None,
        **kwargs,
    ):
        super().__init__(
            nrows=nrows, ncols=ncols, figsize=figsize, style=style, **kwargs
        )
        import seaborn as sns

        self._seaborn = sns

    def _inject_docs(self):
        super()._inject_docs()

        if SeabornFigure._docstring_initialized:
            return

        import seaborn as sns

        for k, f in SeabornFigure.__dict__.items():
            if k.startswith("_") or isinstance(f, property):
                continue
            sns_func = getattr(sns, k, None)
            if sns_func is None:
                continue
            sns_doc = getattr(sns_func, "__doc__", "")
            if sns_doc:
                f.__doc__ = (
                    f"Copy of ``sns.{k}()``. Original docstring " f"is ...\n\n{sns_doc}"
                )

        SeabornFigure._docstring_initialized = True

    @_inject_sns_docs
    def swarmplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.swarmplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def barplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.barplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def boxplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.boxplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def boxenplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.boxenplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def violinplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.violinplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def pointplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.pointplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def histplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.histplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def kdeplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.kdeplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def rugplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.rugplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    def regplot(self, *args, **kwargs) -> Axes:
        out = self._seaborn.regplot(ax=self.ax, *args, **kwargs)
        self.draw()
        return out

    @_inject_sns_docs
    @_use_seaborn_grid
    def jointplot(self, *args, **kwargs) -> sns.JointGrid:
        return self._seaborn.jointplot(*args, **kwargs)

    @_inject_sns_docs
    @_use_seaborn_grid
    def lmplot(self, *args, **kwargs) -> sns.FacetGrid:
        return self._seaborn.lmplot(*args, **kwargs)

    @_inject_sns_docs
    @_use_seaborn_grid
    def pairplot(self, *args, **kwargs) -> sns.PairGrid:
        return self._seaborn.pairplot(*args, **kwargs)



from psygnal import Signal
from ...widgets.utils import FreeWidget


class QtConsole(FreeWidget):
    executed = Signal(str)

    def __init__(self, **kwargs):
        from ._qt import _Console

        super().__init__(**kwargs)

        self.console = _Console()
        self.set_widget(self.console)

    @property
    def __magicclass_parent__(self):
        """Return the parent UI object."""
        return self.console.parent_ui

    @__magicclass_parent__.setter
    def __magicclass_parent__(self, parent):
        while parent.__magicclass_parent__ is not None:
            parent = parent.__magicclass_parent__
        self.console.connect_parent(parent)

    @property
    def value(self) -> str:
        """Get current code block"""
        return self.console.input_buffer

    @value.setter
    def value(self, code: str) -> None:
        """Set code string to Jupyter QtConsole buffer"""
        self.console.input_buffer = ""
        if not isinstance(code, str):
            raise ValueError(f"Cannot set {type(code)}.")
        cursor = self.console._control.textCursor()
        lines = code.split("\n")
        for line in lines[:-1]:
            cursor.insertText(line + "\n")
            self.console._insert_continuation_prompt(cursor)  # insert "...:"
        cursor.insertText(lines[-1])

    def execute(self):
        """Execute current code block."""
        code = self.value
        self.console.execute()
        self.executed.emit(code)



from ...widgets.utils import FreeWidget


class NapariCanvas(FreeWidget):
    def __init__(self, ndisplay=2, order=(), axis_labels=(), **kwargs):
        import napari

        super().__init__(layout="grid", *kwargs)
        self.viewer = napari.Viewer(
            ndisplay=ndisplay, order=order, axis_labels=axis_labels, show=False
        )

        window = self.viewer.window._qt_viewer.parent().parent()
        self.native.setStyleSheet(window.styleSheet())
        window.setParent(self.native, window.windowFlags())

        # _canvas_overlay is the central QWidget that display layers.
        canvas = self.viewer.window._qt_viewer._canvas_overlay
        canvas.setMinimumHeight(200)
        canvas.setMinimumWidth(200)
        self.set_widget(canvas)

        # dims is the multi-slider widget.
        slider = self.viewer.window._qt_viewer.dims
        self.set_widget(slider)

        # layer control widget should be accessible in GUI, but hidden by default.
        self.layer_controls = FreeWidget(name="layer controls")
        self.layer_controls.set_widget(self.viewer.window._qt_viewer.controls)

        self.layer_controls.native.setParent(
            self.viewer.window._qt_viewer, self.layer_controls.native.windowFlags()
        )

        # layer list widget should be accessible in GUI, but hidden by default.
        self.layer_list = FreeWidget(name="layer list")
        self.layer_list.native.setStyleSheet(window.styleSheet())
        self.layer_list.set_widget(self.viewer.window._qt_viewer.dockLayerList.widget())

        self.layer_list.native.setParent(
            self.viewer.window._qt_viewer, self.layer_list.native.windowFlags()
        )



from __future__ import annotations
from typing import Any
import pandas as pd
from pandas._typing import Axes, Dtype, Axis
from magicgui.widgets import Table

from ...fields import field

Defaults = {
    "read_only": True,
}


class WidgetSeries(pd.Series):
    _table = field(Table)

    @property
    def _constructor(self):
        return self.__class__

    @property
    def _constructor_expanddim(self):
        return WidgetDataFrame

    def __init__(
        self,
        data=None,
        index=None,
        dtype: Dtype | None = None,
        name=None,
        copy: bool = False,
        fastpath: bool = False,
    ):
        super().__init__(
            data=data, index=index, dtype=dtype, name=name, copy=copy, fastpath=fastpath
        )
        self._table_initialized = False

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self._table_initialized:
            with self._table.changed.blocked():
                self._table[key] = value

    @_table.connect
    def _on_table_data_change(self, info: dict[str, Any]):
        nc = info["column"]
        if nc > 0:
            raise ValueError(f"Cannot set value at column-{nc}")
        self.iloc[info["row"]] = info["data"]

    @property
    def table(self) -> Table:
        tab = self._table
        with tab.changed.blocked():
            tab.value = {self.name: self.values}
            tab.row_headers = self.index
        tab.read_only = Defaults["read_only"]
        self._table_initialized = True
        return tab


class WidgetDataFrame(pd.DataFrame):
    _table = field(Table)

    @property
    def _constructor(self):
        return self.__class__

    _constructor_sliced = WidgetSeries

    def __init__(
        self,
        data=None,
        index: Axes | None = None,
        columns: Axes | None = None,
        dtype: Dtype | None = None,
        copy: bool | None = None,
    ):
        super().__init__(
            data=data, index=index, columns=columns, dtype=dtype, copy=copy
        )
        self._table_initialized = False

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self._table_initialized:
            with self._table.changed.blocked():
                self._table[key] = value

    @_table.connect
    def _on_table_data_change(self, info: dict[str, Any]):
        self.iloc[info["row"], info["column"]] = info["data"]

    def aggregate(self, func=None, axis: Axis = 0, *args, **kwargs):
        out = super().aggregate(func, axis, *args, **kwargs)
        return self.__class__(out)

    agg = aggregate

    @property
    def table(self) -> Table:
        tab = self._table
        with tab.changed.blocked():
            tab.value = self
        self._table_initialized = True
        tab.read_only = Defaults["read_only"]
        return tab

    _dataframe_index = pd.DataFrame.index
    _dataframe_columns = pd.DataFrame.columns

    @property
    def index(self) -> pd.Index:
        return self._dataframe_index

    @index.setter
    def index(self, value) -> None:
        self._dataframe_index = value
        if self._table_initialized:
            self._table.row_headers = value

    @property
    def columns(self) -> pd.Index:
        return self._dataframe_columns

    @columns.setter
    def columns(self, value) -> None:
        self._dataframe_columns = value
        if self._table_initialized:
            self._table.column_headers = value



from ._dataframe import WidgetDataFrame as DataFrame, WidgetSeries as Series, Defaults
from .io import read_excel, read_csv

__all__ = [
    "DataFrame",
    "Series",
    "read_excel",
    "read_csv",
    "Defaults",
]



from __future__ import annotations
from typing import Sequence
import vtk
import vedo
from vedo.utils import numpy2vtk
import numpy as np
from magicgui.widgets import FloatSlider

from .components import VtkComponent
from .const import Mode, Rendering

from ...fields import vfield
from ...widgets import FloatRangeSlider
from ...types import Color


def split_rgba(col: str | Sequence[float]) -> tuple[str | Sequence[float], float]:
    if not isinstance(col, str):
        if len(col) == 3:
            rgb, alpha = col, 255
        else:
            rgb, alpha = col[:3], col[3]
    elif col.startswith("#"):
        l = len(col)
        if l == 9:
            rgb = int(col[1:7], base=16) / 255
            alpha = int(col[7:], base=16) / 255
        elif l == 7:
            rgb = int(col[1:7], base=16) / 255
            alpha = 1.0
        else:
            raise ValueError(f"Informal color code: {col}.")
    else:
        rgb, alpha = col, 255
    return [int(c * 255) for c in rgb], alpha * 255


class Volume(VtkComponent, base=vedo.Volume):
    _obj: vedo.Volume

    def __init__(self, data, _parent):
        super().__init__(data, _parent=_parent, _emit=False)
        self._current_obj = self._obj
        self.rendering = Rendering.composite
        self.mode = Mode.volume
        self.color = np.array([0.7, 0.7, 0.7])
        self.data = data
        self.contrast_limits = self._lims
        self.iso_threshold = np.mean(self._lims)
        self.widgets.emit_all()

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, v):
        self._data = np.asarray(v)
        self._cache_lims()
        vimg = vtk.vtkImageData()
        varr = numpy2vtk(self._data.ravel(order="F"), dtype=float)
        varr.SetName("input_scalars")
        vimg.SetDimensions(self._data.shape)
        vimg.GetPointData().AddArray(varr)
        vimg.GetPointData().SetActiveScalars(varr.GetName())
        self._obj._update(vimg)
        self._update_actor()

    def _cache_lims(self):
        self._lims = self._data.min(), self._data.max()
        self.widgets.contrast_limits.min = self._lims[0]
        self.widgets.contrast_limits.max = self._lims[1]
        self.widgets.iso_threshold.min = self._lims[0]
        self.widgets.iso_threshold.max = self._lims[1]

    color = vfield(Color)
    mode = vfield(Mode.volume)
    rendering = vfield(Rendering.mip)
    iso_threshold = vfield(float, widget_type=FloatSlider)
    contrast_limits = vfield(tuple[float, float], widget_type=FloatRangeSlider)
    visible = vfield(True, name="visibility")

    @visible.connect
    def _on_visible_change(self, v: bool):
        if v:
            self._current_obj.on()
        else:
            self._current_obj.off()
        self._update()

    @color.connect
    def _on_color_change(self, col):
        rgb, alpha = split_rgba(col)
        vmin, vmax = self.contrast_limits
        self._obj.color(rgb, vmin=vmin, vmax=vmax)
        self._obj.alpha([0, alpha])
        self._update_cmap()

    @rendering.connect
    def _on_rendering_change(self, v: Rendering):
        self._obj.mode(v.value)
        self._update_cmap()

    @iso_threshold.connect
    def _on_iso_threshold_change(self, v):
        if self.mode in (Mode.iso, Mode.wireframe):
            self._update_actor()

    @mode.connect
    def _on_mode_change(self, v):
        self._update_actor()

    @contrast_limits.connect
    def _on_contrast_limits_change(self, v):
        self._update_cmap()

    def _update_cmap(self):
        vmin, vmax = self.contrast_limits
        rgb, alpha = split_rgba(self.color)
        if self.mode == Mode.volume:
            self._current_obj.color(
                [[0, 0, 0], rgb],
                alpha=[0, alpha],
                vmin=vmin,
                vmax=vmax,
            )
        else:
            self._current_obj.color(
                rgb,
                alpha=alpha,
            )
        self._update()

    def _update_actor(self):
        vmin, vmax = self.contrast_limits
        rgb, alpha = split_rgba(self.color)
        if self.mode == Mode.volume:
            actor = self._obj
        elif self.mode == Mode.mesh:
            actor = self._obj.tomesh()
        elif self.mode == Mode.iso:
            actor = self._obj.isosurface(threshold=self.iso_threshold)
        elif self.mode == Mode.wireframe:
            actor = (
                self._obj.isosurface(threshold=self.iso_threshold)
                .color(rgb, alpha=alpha)
                .wireframe()
            )
        elif self.mode == Mode.lego:
            actor = self._obj.legosurface(vmin=vmin, vmax=vmax)
        else:
            raise RuntimeError()

        plotter = self._parent_ref()
        plotter.remove(self._current_obj)
        plotter.add(actor)
        plotter.window.Render()
        self._current_obj = actor
        self._update_cmap()
        return None



from magicclass import magicclass, set_options, field
from magicclass import widgets as wdt

def test_checkbutton():
    @magicclass
    class A:
        b = field(wdt.CheckButton)
        @set_options(x={"widget_type": wdt.CheckButton})
        def f(self, x: bool):
            pass
    ui = A()
    assert isinstance(ui.b, wdt.CheckButton)
    ui.b.value = True
    assert ui.b.native.isChecked()
    ui.b.value = False
    assert not ui.b.native.isChecked()

    ui["f"].changed()
    assert isinstance(ui["f"].mgui[0], wdt.CheckButton)

# def test_figure():
#     @magicclass
#     class A:
#         plt = field(wdt.Figure)

#     ui = A()
#     ui.plt.plot([1,2,3,4])
#     ui.plt.scatter([1,2,3,4], [1,2,3,4])
#     ui.plt.xlim()
#     ui.plt.xlim(0, 2)
#     ui.plt.ylim()
#     ui.plt.ylim(0, 2)
#     ui.plt.cla()
#     ui.plt.title("title")
#     ui.plt.xlabel("x")
#     ui.plt.ylabel("y")
#     ui.plt.imshow([[0, 0], [1, 1]])
#     ui.plt.legend()



from __future__ import annotations
from typing import Sequence
import pyqtgraph as pg
from qtpy.QtCore import Qt
import numpy as np

from .._shared_utils import convert_color_code, to_rgba

# compatibility with matplotlib
_LINE_STYLE = {
    "-": Qt.SolidLine,
    "--": Qt.DashLine,
    ":": Qt.DotLine,
    "-.": Qt.DashDotLine,
}

_SYMBOL_MAP = {
    "*": "star",
    "D": "d",
    "^": "t1",
    "<": "t3",
    "v": "t",
    ">": "t2",
}


class LayerItem:
    native: pg.GraphicsItem

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} '{self.name}'"

    @property
    def visible(self):
        """Visibility of data."""
        return self.native.isVisible()

    @visible.setter
    def visible(self, value: bool):
        self.native.setVisible(value)

    @property
    def zorder(self) -> float:
        """Z-order of item. Item with larger z will be displayed on the top."""
        return self.native.zValue()

    @zorder.setter
    def zorder(self, value: float):
        self.native.setZValue(value)


class PlotDataLayer(LayerItem):
    native: pg.PlotCurveItem | pg.ScatterPlotItem

    @property
    def xdata(self) -> np.ndarray:
        return self.native.getData()[0]

    @xdata.setter
    def xdata(self, value: Sequence[float]):
        self.native.setData(value, self.ydata)

    @property
    def ydata(self) -> np.ndarray:
        return self.native.getData()[1]

    @ydata.setter
    def ydata(self, value: Sequence[float]):
        self.native.setData(self.xdata, value)

    @property
    def ndata(self) -> int:
        return self.native.getData()[0].size

    @property
    def name(self) -> str:
        return self.native.opts["name"]

    @name.setter
    def name(self, value: str):
        value = str(value)
        self.native.opts["name"] = value
        # TODO: now name is not linked to label item

    def add(self, points: np.ndarray | Sequence, **kwargs):
        """Add new points to the plot data item."""
        points = np.atleast_2d(points)
        if points.shape[1] != 2:
            raise ValueError("Points must be of the shape (N, 2).")
        self.native.setData(
            np.concatenate([self.xdata, points[:, 0]]),
            np.concatenate([self.ydata, points[:, 1]]),
            **kwargs,
        )
        return None

    def remove(self, i: int | Sequence[int]):
        """Remove the i-th data."""
        if isinstance(i, int):
            i = [i]
        sl = list(set(range(self.ndata)) - set(i))
        xdata = self.xdata[sl]
        ydata = self.ydata[sl]
        self.native.setData(xdata, ydata)
        return None

    @property
    def edge_color(self) -> np.ndarray:
        """Edge color of the data."""
        return to_rgba(self.native.opts["pen"])

    @edge_color.setter
    def edge_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setPen(value, width=self.lw, style=self.ls)

    @property
    def face_color(self) -> np.ndarray:
        """Face color of the data."""
        return to_rgba(self.native.opts["brush"])

    @face_color.setter
    def face_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setBrush(value)

    color = property()

    @color.setter
    def color(self, value: str | Sequence):
        """Set face color and edge color at the same time."""
        self.face_color = value
        self.edge_color = value

    @property
    def lw(self):
        """Line width."""
        return self.native.opts["pen"].width()

    @lw.setter
    def lw(self, value: float):
        self.native.opts["pen"].setWidth(value)

    linewidth = lw  # alias

    @property
    def ls(self):
        """Line style."""
        return self.native.opts["pen"].style()

    @ls.setter
    def ls(self, value: str):
        _ls = _LINE_STYLE[value]
        self.native.opts["pen"].setStyle(_ls)

    linestyle = ls  # alias


class Scatter(PlotDataLayer):
    native: pg.ScatterPlotItem

    def __init__(
        self,
        x,
        y,
        face_color=None,
        edge_color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        symbol="o",
    ):
        face_color, edge_color = _set_default_colors(
            face_color, edge_color, "white", "white"
        )
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        brush = pg.mkBrush(face_color)
        symbol = _SYMBOL_MAP.get(symbol, symbol)
        self.native = pg.ScatterPlotItem(
            x=x, y=y, pen=pen, brush=brush, size=size, symbol=symbol
        )
        self.name = name

    @property
    def symbol(self):
        return self.native.opts["symbol"]

    @symbol.setter
    def symbol(self, value):
        value = _SYMBOL_MAP.get(value, value)
        self.native.setSymbol(value)

    @property
    def size(self):
        return self.native.opts["symbolSize"]

    @size.setter
    def size(self, size: float):
        self.native.setSymbolSize(size)


class Curve(PlotDataLayer):
    native: pg.PlotDataItem

    def __init__(
        self,
        x,
        y,
        face_color=None,
        edge_color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        symbol=None,
    ):
        face_color, edge_color = _set_default_colors(
            face_color, edge_color, "white", "white"
        )
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        brush = pg.mkBrush(face_color)
        symbol = _SYMBOL_MAP.get(symbol, symbol)
        self.native = pg.PlotDataItem(
            x=x,
            y=y,
            pen=pen,
            brush=brush,
            symbolSize=size,
            symbol=symbol,
            symbolPen=pen,
            symbolBrush=brush,
        )
        self.name = name

    @property
    def symbol(self):
        return self.native.opts["symbol"]

    @symbol.setter
    def symbol(self, value):
        value = _SYMBOL_MAP.get(value, value)
        self.native.setSymbol(value)

    @property
    def size(self):
        return self.native.opts["symbolSize"]

    @size.setter
    def size(self, size: float):
        self.native.setSymbolSize(size)

    @property
    def edge_color(self) -> np.ndarray:
        rgba = self.native.opts["pen"].color().getRgb()
        return np.array(rgba) / 255

    @edge_color.setter
    def edge_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setPen(value)
        self.native.setSymbolPen(value)

    @property
    def face_color(self) -> np.ndarray:
        rgba = self.native.opts["symbolBrush"].color().getRgb()
        return np.array(rgba) / 255

    @face_color.setter
    def face_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setBrush(value)
        self.native.setSymbolBrush(value)


class Histogram(PlotDataLayer):
    native: pg.ScatterPlotItem

    def __init__(
        self,
        data,
        bins: int | Sequence | str = 10,
        range=None,
        density: bool = False,
        face_color=None,
        edge_color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        face_color, edge_color = _set_default_colors(
            face_color, edge_color, "white", "white"
        )
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        brush = pg.mkBrush(face_color)
        y, x = np.histogram(data, bins=bins, range=range, density=density)
        self.native = pg.PlotCurveItem(
            x=x, y=y, pen=pen, brush=brush, stepMode="center", fillLevel=0
        )
        self.name = name

    def set_hist(self, data, bins=10, range=None, density=False):
        y, x = np.histogram(data, bins=bins, range=range, density=density)
        self.native.setData(x=x, y=y)


class BarPlot(PlotDataLayer):
    native: pg.BarGraphItem

    def __init__(
        self,
        x,
        y,
        face_color=None,
        edge_color=None,
        width: float = 0.6,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        face_color, edge_color = _set_default_colors(
            face_color, edge_color, "white", "white"
        )
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        brush = pg.mkBrush(face_color)
        self.native = pg.BarGraphItem(x=x, height=y, width=width, pen=pen, brush=brush)
        self.name = name

    @property
    def edge_color(self) -> np.ndarray:
        rgba = self.native.opts["pen"].color().getRgb()
        return np.array(rgba) / 255

    @edge_color.setter
    def edge_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setOpts(pen=pg.mkPen(value))

    @property
    def face_color(self) -> np.ndarray:
        rgba = self.native.opts["brush"].color().getRgb()
        return np.array(rgba) / 255

    @face_color.setter
    def face_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setOpts(brush=pg.mkBrush(value))

    @property
    def xdata(self) -> np.ndarray:
        return self.native.getData()[0]

    @xdata.setter
    def xdata(self, value: Sequence[float]):
        self.native.setOpts(x=value)

    @property
    def ydata(self) -> np.ndarray:
        return self.native.getData()[1]

    @ydata.setter
    def ydata(self, value: Sequence[float]):
        self.native.setOpts(height=value)


class InfLine(LayerItem):
    native: pg.InfiniteLine

    def __init__(
        self,
        pos,
        angle,
        edge_color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        if edge_color is None:
            edge_color = "yellow"
        edge_color = convert_color_code(edge_color)
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        self.native = pg.InfiniteLine(pos, angle, pen=pen, name=name)
        self.name = name

    @property
    def slope(self) -> float:
        """Slope of the line."""
        return np.tan(np.deg2rad(self.native.angle))

    @slope.setter
    def slope(self, value: float):
        self.native.setAngle(np.rad2deg(np.arctan(value)))

    @property
    def intercept(self) -> float:
        """Y-intercept of the line."""
        a = self.slope
        x0, y0 = self.native.getPos()
        return y0 - a * x0

    @intercept.setter
    def intercept(self, value: float):
        value = float(value)
        self.native.setPos((0, value))

    @property
    def pos(self) -> np.ndarray:
        return np.array(self.native.getPos())

    @pos.setter
    def pos(self, value):
        self.native.setPos(value)

    @property
    def angle(self) -> float:
        """Angle of the line in degree."""
        return self.native.angle

    @angle.setter
    def angle(self, value: float):
        self.native.setAngle(value)

    @property
    def edge_color(self) -> np.ndarray:
        return to_rgba(self.native.pen)

    @edge_color.setter
    def edge_color(self, value):
        value = convert_color_code(value)
        self.native.setPen(value, width=self.lw, style=self.ls)

    color = edge_color

    @property
    def name(self):
        return self.native._name

    @name.setter
    def name(self, value: str):
        self.native.setName(value)

    @property
    def lw(self):
        """Line width."""
        return self.native.pen.width()

    @lw.setter
    def lw(self, value: float):
        self.native.pen.setWidth(value)

    linewidth = lw  # alias

    @property
    def ls(self):
        """Line style."""
        return self.native.pen.style()

    @ls.setter
    def ls(self, value: str):
        _ls = _LINE_STYLE[value]
        self.native.pen.setStyle(_ls)

    linestyle = ls  # alias


class FillBetween(PlotDataLayer):
    native: pg.FillBetweenItem

    def __init__(
        self,
        x,
        y1,
        y2,
        face_color=None,
        edge_color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        face_color, edge_color = _set_default_colors(
            face_color, edge_color, "white", "white"
        )
        pen = pg.mkPen(edge_color, width=lw, style=_LINE_STYLE[ls])
        brush = pg.mkBrush(face_color)
        curve1 = pg.PlotCurveItem(x=x, y=y1, pen=pen)
        curve2 = pg.PlotCurveItem(x=x, y=y2, pen=pen)
        self.native = pg.FillBetweenItem(curve1, curve2, brush=brush, pen=pen)
        self.name = name

    @property
    def edge_color(self) -> np.ndarray:
        rgba = self.native.curves[0].opts["pen"].color().getRgb()
        return np.array(rgba) / 255

    @edge_color.setter
    def edge_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setPen(pg.mkPen(value))

    @property
    def face_color(self) -> np.ndarray:
        rgba = self.native.curves[0].opts["brush"].color().getRgb()
        return np.array(rgba) / 255

    @face_color.setter
    def face_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setBrush(pg.mkBrush(value))

    @property
    def name(self) -> str:
        return self.native.curves[0].opts["name"]

    @name.setter
    def name(self, value: str):
        value = str(value)
        self.native.curves[0].opts["name"] = value

    @property
    def lw(self):
        """Line width."""
        return self.native.curves[0].opts["pen"].width()

    @lw.setter
    def lw(self, value: float):
        self.native.curves[0].opts["pen"].setWidth(value)
        self.native.curves[1].opts["pen"].setWidth(value)

    linewidth = lw  # alias

    @property
    def ls(self):
        """Line style."""
        return self.native.curves[0].opts["pen"].style()

    @ls.setter
    def ls(self, value: str):
        _ls = _LINE_STYLE[value]
        self.native.curves[0].opts["pen"].setStyle(_ls)
        self.native.curves[1].opts["pen"].setStyle(_ls)

    linestyle = ls  # alias


# WIP!
# How to update a subset of properties? item.text[2:5] = "new" or item[2:5].text = "new"
class TextGroup(LayerItem):
    def __init__(
        self,
        x: Sequence[float],
        y: Sequence[float],
        texts: Sequence[str],
        color=None,
        name: str = None,
    ):
        self.native = pg.ItemGroup()
        if color is None:
            color = "white"
        for x_, y_, text_ in zip(x, y, texts):
            item = pg.TextItem(text_, color=convert_color_code(color))
            item.setPos(x_, y_)
            self.native.addItem(item)

        self.name = name

    @property
    def text_items(self) -> list[pg.TextItem]:
        return self.native.childItems()

    def __getitem__(self, key: int | slice) -> TextItemView:
        return TextItemView(self.text_items[key])

    @property
    def xdata(self) -> np.ndarray:
        return np.array([item.pos().x() for item in self.text_items])

    @property
    def ydata(self) -> np.ndarray:
        return np.array([item.pos().y() for item in self.text_items])

    @property
    def color(self) -> np.ndarray:
        """Text color."""
        rgba = np.stack([item.color.getRgb() for item in self.text_items])
        return rgba / 255

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        for item in self.text_items:
            item.setText(item.toPlainText(), value)

    @property
    def background_color(self) -> np.ndarray:
        """Text background color."""
        return np.stack([to_rgba(item.fill) for item in self.text_items])

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        brush = pg.mkBrush(value)
        for item in self.text_items:
            item.fill = brush
            item._updateView()

    @property
    def border(self) -> np.ndarray:
        """Border color of text bounding box."""
        if isinstance(self.native, list):
            return np.stack([to_rgba(item.border) for item in self.native])
        else:
            return to_rgba(self.native.border)

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        pen = pg.mkPen(value)

        for item in self.text_items:
            item.border = pen
            item._updateView()

    @property
    def text(self) -> str | list[str]:
        """Text string."""
        return [item.toPlainText() for item in self.text_items]

    @text.setter
    def text(self, value: str):
        for item in self.text_items:
            item.setText(value)

    @property
    def anchor(self) -> np.ndarray:
        """Text anchor position."""
        out = []
        for item in self.text_items:
            anchor = item.anchor
            out.append([anchor.x(), anchor.y()])
        return np.array(out)

    @anchor.setter
    def anchor(self, value):
        for item in self.text_items:
            item.setAnchor(value)


class TextItemView:
    def __init__(self, textitem: pg.TextItem | list[pg.TextItem]):
        self.native = textitem

    @property
    def color(self) -> np.ndarray:
        """Text color."""
        if isinstance(self.native, list):
            rgba = np.stack([item.color.getRgb() for item in self.native])
            arr = rgba / 255
        else:
            rgba = self.native.color.getRgb()
            arr = np.array(rgba) / 255
        return arr

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        if isinstance(self.native, list):
            for item in self.native:
                item.setText(item.text, value)
        else:
            self.native.setText(self.text, value)

    @property
    def background_color(self) -> np.ndarray:
        """Text background color."""
        if isinstance(self.native, list):
            return np.stack([to_rgba(item.fill) for item in self.native])
        else:
            return to_rgba(self.native.fill)

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        brush = pg.mkBrush(value)
        if isinstance(self.native, list):
            for item in self.native:
                item.fill = brush
                item._updateView()
        else:
            self.native.fill = brush
            self.native._updateView()

    @property
    def border(self) -> np.ndarray:
        """Border color of text bounding box."""
        if isinstance(self.native, list):
            return np.stack([to_rgba(item.border) for item in self.native])
        else:
            return to_rgba(self.native.border)

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        pen = pg.mkPen(value)

        if isinstance(self.native, list):
            for item in self.native:
                item.border = pen
                item._updateView()
        else:
            self.native.border = pen
            self.native._updateView()

    @property
    def text(self) -> str | list[str]:
        """Text string."""
        if isinstance(self.native, list):
            return [item.toPlainText() for item in self.native]
        else:
            return self.native.toPlainText()

    @text.setter
    def text(self, value: str):

        if isinstance(self.native, list):
            for item in self.native:
                item.setText(value)
        else:
            self.native.setText(value)

    @property
    def anchor(self) -> np.ndarray:
        """Text anchor position."""
        if isinstance(self.native, list):
            out = []
            for item in self.native:
                anchor = item.anchor
                out.append([anchor.x(), anchor.y()])
            return np.array(out)
        else:
            anchor = self.native.anchor
            return np.array([anchor.x(), anchor.y()])

    @anchor.setter
    def anchor(self, value):
        if isinstance(self.native, list):
            for item in self.native:
                item.setAnchor(value)
        else:
            self.native.setAnchor(value)


def _set_default_colors(face_color, edge_color, default_f, default_e):
    if face_color is None:
        face_color = default_f
    else:
        face_color = convert_color_code(face_color)
    if edge_color is None:
        edge_color = default_e
    else:
        edge_color = convert_color_code(edge_color)
    return face_color, edge_color



from __future__ import annotations
from typing import Iterable
from qtpy.QtWidgets import (
    QLineEdit,
    QColorDialog,
    QFrame,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QDoubleSpinBox,
    QSlider,
)
from qtpy.QtGui import QColor
from qtpy.QtCore import Qt, Signal as QtSignal

from magicgui.widgets._bases.value_widget import ValueWidget
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from magicgui.application import use_app
from .utils import merge_super_sigs


def rgba_to_qcolor(rgba: Iterable[float]) -> QColor:
    return QColor(*[int(round(255 * c)) for c in rgba])


def qcolor_to_rgba(qcolor: QColor) -> tuple[float, float, float, float]:
    return tuple(c / 255 for c in qcolor.getRgb())


def rgba_to_html(rgba: Iterable[float]) -> str:
    code = "#" + "".join(hex(int(c * 255))[2:].upper().zfill(2) for c in rgba)
    if code.endswith("FF"):
        code = code[:-2]
    return code


# modified from napari/_qt/widgets/qt_color_swatch.py
class QColorSwatch(QFrame):
    colorChanged = QtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCursor(Qt.PointingHandCursor)
        self._color: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
        self.colorChanged.connect(self._update_swatch_style)
        self.setMinimumWidth(40)

    def heightForWidth(self, w: int) -> int:
        return int(w * 0.667)

    def _update_swatch_style(self, _=None) -> None:
        rgba = f'rgba({",".join(str(int(x*255)) for x in self._color)})'
        self.setStyleSheet("QColorSwatch {background-color: " + rgba + ";}")

    def mouseReleaseEvent(self, event):
        """Show QColorPopup picker when the user clicks on the swatch."""
        if event.button() == Qt.LeftButton:
            initial = self.getQColor()
            dlg = QColorDialog(initial, self)
            dlg.setOptions(QColorDialog.ShowAlphaChannel)
            ok = dlg.exec_()
            if ok:
                self.setColor(dlg.selectedColor())

    def getQColor(self) -> QColor:
        return rgba_to_qcolor(self._color)

    def setColor(self, color: QColor) -> None:
        old_color = rgba_to_html(self._color)
        self._color = qcolor_to_rgba(color)
        if rgba_to_html(self._color) != old_color:
            self.colorChanged.emit()


class QColorLineEdit(QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        import matplotlib.colors

        self._color_converter = matplotlib.colors.to_rgba

    def setText(self, color: str | Iterable[float]):
        """Set the text of the lineEdit using any ColorType.

        Colors will be converted to standard SVG spec names if possible,
        or shown as #RGBA hex if not.

        Parameters
        ----------
        color : ColorType
            Can be any ColorType recognized by our
            utils.colormaps.standardize_color.transform_color function.
        """
        if isinstance(color, QColor):
            color = rgba_to_html(qcolor_to_rgba(color))
        elif not isinstance(color, str):
            color = rgba_to_html(color)

        super().setText(color)

    def getQColor(self) -> QColor:
        """Get color as QColor object"""
        rgba = self._color_converter(self.text())
        return rgba_to_qcolor(rgba)

    def setColor(self, color: QColor | str):
        if isinstance(color, str):
            color = self._color_converter(color)
        elif isinstance(color, QColor):
            color = qcolor_to_rgba(color)
        code = "#" + "".join(
            hex(int(round(c * 255)))[2:].upper().zfill(2) for c in color
        )
        if code.endswith("FF"):
            code = code[:-2]
        self.setText(code)


class QColorEdit(QWidget):
    colorChanged = QtSignal(tuple)

    def __init__(self, parent=None, value: str = "white"):
        super().__init__(parent)
        _layout = QHBoxLayout()
        self._color_swatch = QColorSwatch(self)
        self._line_edit = QColorLineEdit(self)
        _layout.addWidget(self._color_swatch)
        _layout.addWidget(self._line_edit)
        self.setLayout(_layout)

        self._color_swatch.colorChanged.connect(self._on_swatch_changed)
        self._line_edit.editingFinished.connect(self._on_line_edit_edited)

        self._line_edit.setColor(value)
        self._color_swatch.setColor(self._line_edit.getQColor())

    def color(self):
        """Return the current color."""
        return self._color_swatch._color

    def setColor(self, color):
        """Set value as the current color."""
        if isinstance(color, QColor):
            color = qcolor_to_rgba(color)
        self._line_edit.setText(color)
        color = self._line_edit.getQColor()
        self._color_swatch.setColor(color)

    def _on_line_edit_edited(self, _=None):
        text = self._line_edit.text()
        try:
            self._line_edit.setText(text)
        except ValueError:
            self._on_swatch_changed()
        else:
            qcolor = self._line_edit.getQColor()
            self._color_swatch.setColor(qcolor)

    def _on_swatch_changed(self, _=None):
        qcolor = self._color_swatch.getQColor()
        self._line_edit.setColor(qcolor)
        self.colorChanged.emit(self.color())


# See https://stackoverflow.com/questions/42820380/use-float-for-qslider
class QDoubleSlider(QSlider):
    changed = QtSignal(float)

    def __init__(self, parent=None, decimals: int = 3):
        super().__init__(parent=parent)
        self.scale = 10 ** decimals
        self.setOrientation(Qt.Horizontal)
        self.valueChanged.connect(self.doubleValueChanged)

    def doubleValueChanged(self):
        value = float(super().value()) / self.scale
        self.changed.emit(value)

    def value(self):
        return float(super().value()) / self.scale

    def setValue(self, value):
        super().setValue(int(float(value) * self.scale))

    def setMinimum(self, value):
        return super().setMinimum(value * self.scale)

    def setMaximum(self, value):
        return super().setMaximum(value * self.scale)

    def singleStep(self):
        return float(super().singleStep()) / self.scale

    def setSingleStep(self, value):
        return super().setSingleStep(value * self.scale)


class QColorSlider(QWidget):
    colorChanged = QtSignal(tuple)

    def __init__(self, parent=None, value="white"):
        super().__init__(parent=parent)
        import matplotlib.colors

        self._color_converter = matplotlib.colors.to_rgba
        _layout = QVBoxLayout()
        self.setLayout(_layout)
        _layout.setContentsMargins(0, 0, 0, 0)
        self._qsliders = [
            self.addSlider("R"),
            self.addSlider("G"),
            self.addSlider("B"),
            self.addSlider("A"),
        ]

        self._color_edit = QColorEdit(self, value=value)

        @self._color_edit._line_edit.editingFinished.connect
        def _read_color_str(e=None):
            self.setColor(self._color_edit._line_edit.getQColor())

        self._color_edit._color_swatch.setEnabled(False)

        @self.colorChanged.connect
        def _set_color_swatch(color: QColor):
            qcolor = rgba_to_qcolor(color)
            self._color_edit._color_swatch.setColor(qcolor)

        _layout.addWidget(self._color_edit)

    def addSlider(self, label: str):
        qlabel = QLabel(label)
        qlabel.setFixedWidth(15)
        qslider = QDoubleSlider()
        qslider.setMaximum(1.0)
        qslider.setSingleStep(0.001)
        qspinbox = QDoubleSpinBox()
        qspinbox.setMaximum(1.0)
        qspinbox.setSingleStep(0.001)
        qspinbox.setAlignment(Qt.AlignRight)

        qspinbox.setButtonSymbols(QDoubleSpinBox.ButtonSymbols.NoButtons)
        qspinbox.setStyleSheet("background:transparent; border: 0;")

        qslider.changed.connect(qspinbox.setValue)
        qslider.changed.connect(lambda e: self.setColor(self.color()))
        qspinbox.editingFinished.connect(qslider.setValue)
        qspinbox.editingFinished.connect(lambda e: self.setColor(qspinbox.text()))

        _container = QWidget(self)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        _container.setLayout(layout)
        layout.addWidget(qlabel)
        layout.addWidget(qslider)
        layout.addWidget(qspinbox)

        self.layout().addWidget(_container)
        return qslider

    def color(self):
        """Return the current color."""
        return tuple(sl.value() for sl in self._qsliders)

    def setColor(self, color):
        """Set value as the current color."""
        if isinstance(color, QColor):
            color = qcolor_to_rgba(color)
        elif isinstance(color, str):
            color = self._color_converter(color)
        self._color_edit.setColor(color)
        for sl, c in zip(self._qsliders, color):
            sl.setValue(c)
        self.colorChanged.emit(self.color())


class _ColorEdit(QBaseValueWidget):
    _qwidget: QColorEdit

    def __init__(self):
        super().__init__(QColorEdit, "color", "setColor", "colorChanged")


class _ColorSlider(QBaseValueWidget):
    _qwidget: QColorSlider

    def __init__(self):
        super().__init__(QColorSlider, "color", "setColor", "colorChanged")


@merge_super_sigs
class ColorEdit(ValueWidget):
    """
    A widget for editing colors.

    Parameters
    ----------
    value : tuple of float or str
        RGBA color, color code or standard color name.
    """

    def __init__(self, **kwargs):
        app = use_app()
        assert app.native
        kwargs["widget_type"] = _ColorEdit
        super().__init__(**kwargs)


@merge_super_sigs
class ColorSlider(ValueWidget):
    """
    A multi-slider for editing colors.

    Parameters
    ----------
    value : tuple of float or str
        RGBA color, color code or standard color name.
    """

    def __init__(self, **kwargs):
        app = use_app()
        assert app.native
        kwargs["widget_type"] = _ColorSlider
        super().__init__(**kwargs)



try:
    from .widgets import (
        QtPlotCanvas,
        QtMultiPlotCanvas,
        Qt2YPlotCanvas,
        QtImageCanvas,
        QtMultiImageCanvas,
    )

    PYQTGRAPH_AVAILABLE = True

    __all__ = [
        "QtPlotCanvas",
        "QtMultiPlotCanvas",
        "Qt2YPlotCanvas",
        "QtImageCanvas",
        "QtMultiImageCanvas",
        "PYQTGRAPH_AVAILABLE",
    ]
except ImportError:

    PYQTGRAPH_AVAILABLE = False

    __all__ = ["PYQTGRAPH_AVAILABLE"]


def _join(strs):
    strs = [f"``{s}``" for s in strs]
    return ", ".join(strs[:-1]) + f" and {strs[-1]}"


__doc__ = f"""
PyQtGraph wrapper classes. Currently supports {_join(__all__)}.

These classes offer unified API between graphical items and components, similar to those
in ``magicgui`` and ``napari``.

``QtPlotCanvas`` can treat line plot, scatter plot, bar plot and histogram as "layers".

.. code-block:: python

    from magicclass.widgets import QtPlotCanvas

    canvas = QtPlotCanvas()
    canvas.add_curve(np.random.random(100), color="r") # like plt.plot
    canvas.add_scatter(np.random.random(100), ls=":")  # like plt.scatter
    canvas.layers[1].visible = False  # toggle visibility
    canvas.enabled = False            # toggle interactivity
    canvas.show()

Other components such as axis labels, title and linear region have intuitive interface.

.. code-block:: python

    canvas.region.visible = True  # show linear region
    canvas.region.value           # get range
    canvas.xlabel = "time"        # change label text
    canvas.xlim = [0, 1]          # change limits.

``QtImageCanvas`` is also designed in a similar way as ``QtPlotCanvas`` but aims at 2D image
visualization.

.. code-block:: python

    from magicclass.widgets import QtImageCanvas

    canvas = QtImageCanvas()
    canvas.image = np.random.random((128, 128))
    canvas.show()

Text overlay and scale bar are available now.

.. code-block:: python

    canvas.text_overlay.text = "some info"
    canvas.text_overlay.color = "lime"
    canvas.scale_bar.unit = "px"
    canvas.scale_bar.visible = True

"""



from __future__ import annotations
from typing import Sequence
import numpy as np
from numpy.typing import ArrayLike
from vispy.scene import visuals, ViewBox
from ._base import LayerItem
from .._shared_utils import convert_color_code, to_rgba

_SYMBOL_MAP = {
    "s": "square",
    "D": "diamond",
}


class PlotDataLayer(LayerItem):
    _visual: visuals.LinePlot | visuals.Markers
    _data: np.ndarray

    @property
    def xdata(self) -> np.ndarray:
        return self._data[:, 0]

    @xdata.setter
    def xdata(self, value: Sequence[float]):

        self._visual.set_data

    @property
    def ydata(self) -> np.ndarray:
        return self._data[:, 1]

    @ydata.setter
    def ydata(self, value: Sequence[float]):
        self.native.setData(self.xdata, value)

    @property
    def ndata(self) -> int:
        return self.xdata.size

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = str(value)

    def add(self, points: np.ndarray | Sequence, **kwargs):
        """Add new points to the plot data item."""
        points = np.atleast_2d(points)
        if points.shape[1] != 2:
            raise ValueError("Points must be of the shape (N, 2).")
        data = np.concatenate([self._data, points], axis=1)
        self._visual.set_data(data)
        return None

    def remove(self, i: int | Sequence[int]):
        """Remove the i-th data."""
        if isinstance(i, int):
            i = [i]
        sl = list(set(range(self.ndata)) - set(i))
        xdata = self.xdata[sl]
        ydata = self.ydata[sl]
        self.native.setData(xdata, ydata)
        return None

    @property
    def edge_color(self) -> np.ndarray:
        """Edge color of the data."""
        return to_rgba(self.native.opts["pen"])

    @edge_color.setter
    def edge_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setPen(value, width=self.lw, style=self.ls)

    @property
    def face_color(self) -> np.ndarray:
        """Face color of the data."""
        return to_rgba(self.native.opts["brush"])

    @face_color.setter
    def face_color(self, value: str | Sequence):
        value = convert_color_code(value)
        self.native.setBrush(value)

    color = property()

    @color.setter
    def color(self, value: str | Sequence):
        """Set face color and edge color at the same time."""
        self.face_color = value
        self.edge_color = value


class Curve(PlotDataLayer):
    def __init__(
        self,
        viewbox: ViewBox,
        x: ArrayLike,
        y: ArrayLike = None,
        face_color=None,
        edge_color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",  # not implemented yet
        symbol=None,
    ) -> None:
        symbol = _SYMBOL_MAP.get(symbol, symbol)
        if symbol is None:
            face_color = None
        self._viewbox = viewbox
        self._visual = visuals.LinePlot(
            np.stack([x, y], axis=1),
            color=edge_color,
            symbol=symbol,
            parent=self._viewbox.scene,
            width=lw,
            marker_size=size,
            face_color=face_color,
            edge_color=face_color,
        )
        self._name = name
        self._visual.update()


class Scatter(PlotDataLayer):
    def __init__(
        self,
        viewbox: ViewBox,
        x: ArrayLike,
        y: ArrayLike = None,
        face_color=None,
        edge_color=None,
        size: float = 7,
        name: str | None = None,
        symbol="o",
    ) -> None:
        symbol = _SYMBOL_MAP.get(symbol, symbol)
        self._viewbox = viewbox
        self._visual = visuals.Markers(
            pos=np.stack([x, y], axis=1),
            symbol=symbol,
            parent=self._viewbox.scene,
            size=size,
            face_color=face_color,
            edge_color=edge_color,
        )
        self._name = name
        self._visual.update()



from __future__ import annotations
import threading
import time
from timeit import default_timer
import inspect
from functools import partial, wraps
from typing import (
    Any,
    Callable,
    TYPE_CHECKING,
    Iterable,
    Union,
    overload,
    TypeVar,
    Generic,
    Protocol,
    runtime_checkable,
)
from typing_extensions import TypedDict, ParamSpec

from superqt.utils import create_worker, GeneratorWorker, FunctionWorker
from qtpy.QtCore import Qt
from magicgui.widgets import ProgressBar, Container, Widget, PushButton, Label
from magicgui.application import use_app

from . import get_signature, move_to_screen_center
from .qtsignal import QtSignal

if TYPE_CHECKING:
    from .._gui import BaseGui
    from .._gui.mgui_ext import PushButtonPlus, Action
    from .._gui._macro import GuiMacro
    from ..fields import MagicField

__all__ = ["thread_worker", "Timer", "Callback"]


class ProgressDict(TypedDict):
    """Supported keys for the progress argument."""

    desc: str | Callable
    total: str | Callable
    pbar: ProgressBar | _SupportProgress | MagicField


@runtime_checkable
class _SupportProgress(Protocol):
    """
    A progress protocol.

    Progressbar must be implemented with methods ``__init__``, ``set_description``,
    ``show``, ``close`` and properties ``value``, ``max``. Optionally ``set_worker``
    can be used so that progressbar has an access to the worker object.
    """

    def __init__(self, max: int = 1, **kwargs):
        raise NotImplementedError()

    @property
    def value(self) -> int:
        raise NotImplementedError()

    @value.setter
    def value(self, v) -> None:
        raise NotImplementedError()

    @property
    def max(self) -> int:
        raise NotImplementedError()

    @max.setter
    def max(self, v) -> None:
        raise NotImplementedError()

    def set_description(self, desc: str):
        raise NotImplementedError()

    def show(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


_P = ParamSpec("_P")
_R1 = TypeVar("_R1")
_R2 = TypeVar("_R2")


class CallbackList(Generic[_R1]):
    """List of callback functions."""

    def __init__(self):
        self._callbacks: list[Callable[[Any, _R1], _R2] | Callable[[Any], _R2]] = []

    @property
    def callbacks(self) -> tuple[Callable[[Any, _R1], _R2] | Callable[[Any], _R2], ...]:
        return tuple(self._callbacks)

    def connect(
        self, callback: Callable[[Any, _R1], _R2] | Callable[[Any], _R2]
    ) -> Callable[[Any, _R1], _R2] | Callable[[Any], _R2]:
        """
        Append a callback function to the callback list.

        Parameters
        ----------
        callback : Callable
            Callback function.
        """
        if not callable(callback):
            raise TypeError("Can only connect callable object.")
        self._callbacks.append(callback)
        return callback

    def disconnect(
        self, callback: Callable[[Any, _R1], _R2] | Callable[[Any], _R2]
    ) -> Callable[[Any, _R1], _R2] | Callable[[Any], _R2]:
        """
        Remove callback function from the callback list.

        Parameters
        ----------
        callback : Callable
            Callback function to be removed.
        """
        self._callbacks.remove(callback)
        return callback

    def _iter_as_method(self, obj: BaseGui) -> Iterable[Callable]:
        for ref in self._callbacks:
            yield _make_method(ref, obj)


def _make_method(func, obj: BaseGui):
    def f(*args, **kwargs):
        with obj.macro.blocked():
            out = func.__get__(obj)(*args, **kwargs)
        return out

    return f


class NapariProgressBar(_SupportProgress):
    """A progressbar class that provides napari progress bar with same API."""

    def __init__(self, value: int = 0, max: int = 1000):
        from napari.utils import progress

        with progress._all_instances.events.changed.blocker():
            self._pbar = progress(total=max)
            self._pbar.n = value

    @property
    def value(self) -> int:
        return self._pbar.n

    @value.setter
    def value(self, v) -> None:
        self._pbar.n = v
        self._pbar.events.value(value=self._pbar.n)

    @property
    def max(self) -> int:
        return self._pbar.total

    @max.setter
    def max(self, v) -> None:
        self._pbar.total = v

    def set_description(self, v: str) -> None:
        self._pbar.set_description(v)

    @property
    def visible(self) -> bool:
        return False

    def show(self):
        type(self._pbar)._all_instances.events.changed(added={self._pbar}, removed={})

    def close(self):
        self._pbar.close()


class Timer:
    """A timer class with intuitive API."""

    def __init__(self):
        self.reset()

    def __repr__(self) -> str:
        """Return string in format hh:mm:ss"""
        return self.format_time()

    @property
    def sec(self) -> float:
        """Return current second."""
        self.lap()
        return self._t_total

    def start(self):
        """Start timer."""
        self._t0 = default_timer()
        self._running = True

    def stop(self) -> float:
        """Stop timer."""
        self.lap()
        self._running = False

    def lap(self) -> float:
        """Return lap time."""
        if self._running:
            now = default_timer()
            self._t_total += now - self._t0
            self._t0 = now
        return self._t_total

    def reset(self):
        """Reset timer."""
        self._t0 = default_timer()
        self._t_total = 0.0
        self._running = False
        return None

    def format_time(self, fmt: str = "{hour:0>2}:{min:0>2}:{sec:0>2}") -> str:
        """Format current time."""
        min_all, sec = divmod(self.sec, 60)
        hour, min = divmod(min_all, 60)
        return fmt.format(hour=int(hour), min=int(min), sec=int(sec))


class DefaultProgressBar(Container, _SupportProgress):
    """The default progressbar widget."""

    def __init__(self, max: int = 1):
        self.progress_label = Label(value="Progress")
        self.pbar = ProgressBar(value=0, max=max)
        self.time_label = Label(value="00:00")
        self.pause_button = PushButton(text="Pause")
        self.abort_button = PushButton(text="Abort")
        cnt = Container(
            layout="horizontal",
            widgets=[self.time_label, self.pause_button, self.abort_button],
            labels=False,
        )
        cnt.margins = (0, 0, 0, 0)
        self.footer = cnt
        self.pbar.min_width = 200
        self._timer = Timer()
        self._time_signal = QtSignal()
        self._time_signal.connect(self._on_timer_updated)

        super().__init__(widgets=[self.progress_label, self.pbar, cnt], labels=False)

    def _on_timer_updated(self, _=None):
        if self._timer.sec < 3600:
            self.time_label.value = self._timer.format_time("{min:0>2}:{sec:0>2}")
        else:
            self.time_label.value = self._timer.format_time()
        return None

    def _start_thread(self):
        # Start background thread
        self._running = True
        self._thread_timer = threading.Thread(target=self._update_timer_label)
        self._thread_timer.daemon = True
        self._thread_timer.start()
        self._timer.start()
        return None

    def _update_timer_label(self):
        """Background thread for updating the progress bar"""
        while self._running:
            if self._timer._running:
                self._time_signal.emit()

            time.sleep(0.1)
        return None

    def _finish(self):
        self._running = False
        self._thread_timer.join()
        return None

    @property
    def paused(self) -> bool:
        return not self._timer._running

    @property
    def value(self) -> int:
        return self.pbar.value

    @value.setter
    def value(self, v):
        self.pbar.value = v

    @property
    def max(self) -> int:
        return self.pbar.max

    @max.setter
    def max(self, v):
        self.pbar.max = v

    def set_description(self, desc: str):
        """Set description as the label of the progressbar."""
        self.progress_label.value = desc
        return None

    def set_worker(self, worker: GeneratorWorker | FunctionWorker):
        """Set currently running worker."""
        self._worker = worker
        self._worker.finished.connect(self._finish)
        self._worker.started.connect(self._start_thread)
        if not isinstance(self._worker, GeneratorWorker):
            # FunctionWorker does not have yielded/aborted signals.
            self.footer[1].visible = False
            self.footer[2].visible = False
            return None
        # initialize abort_button
        self.abort_button.text = "Abort"
        self.abort_button.changed.connect(self._abort_worker)
        self.abort_button.enabled = True

        # initialize pause_button
        self.pause_button.text = "Pause"
        self.pause_button.enabled = True
        self.pause_button.changed.connect(self._toggle_pause)

        @self._worker.paused.connect
        def _on_pause():
            self.pause_button.text = "Resume"
            self.pause_button.enabled = True
            self._timer.stop()

        return None

    def _toggle_pause(self):
        if self.paused:
            self._worker.resume()
            self.pause_button.text = "Pause"
            self._timer.start()
        else:
            self._worker.pause()
            self.pause_button.text = "Pausing"
            self.pause_button.enabled = False

        return None

    def _abort_worker(self):
        self.pause_button.text = "Pause"
        self.abort_button.text = "Aborting"
        self.pause_button.enabled = False
        self.abort_button.enabled = False
        self._worker.quit()
        return None


class Aborted(RuntimeError):
    """Raised when worker is aborted."""

    @classmethod
    def raise_(cls, *args):
        """A function version of "raise"."""
        if not args:
            args = ("Aborted.",)
        raise cls(*args)


ProgressBarLike = Union[ProgressBar, _SupportProgress]


class thread_worker:
    """Create a worker in a superqt/napari style."""

    _DEFAULT_PROGRESS_BAR = DefaultProgressBar
    _DEFAULT_TOTAL = 0
    _WINDOW_FLAG = Qt.WindowTitleHint | Qt.WindowMinimizeButtonHint | Qt.Window

    def __init__(
        self,
        f: Callable[_P, _R1] | None = None,
        *,
        ignore_errors: bool = False,
        progress: ProgressDict | None = None,
    ) -> None:
        self._func: Callable[_P, _R1] | None = None
        self._callback_dict_ = {
            "started": CallbackList(),
            "returned": CallbackList(),
            "errored": CallbackList(),
            "yielded": CallbackList(),
            "finished": CallbackList(),
            "aborted": CallbackList(),
        }

        self._ignore_errors = ignore_errors
        self._objects: dict[int, BaseGui] = {}
        self._progressbars: dict[int, ProgressBarLike | None] = {}
        self._recorder: Callable[_P, Any] | None = None

        if f is not None:
            self(f)

        if progress:
            if isinstance(progress, bool):
                progress = {}

            progress.setdefault("desc", None)
            progress.setdefault("total", 0)
            progress.setdefault("pbar", None)

        self._progress = progress

    @property
    def func(self) -> Callable[_P, _R1]:
        return self._func

    @classmethod
    def set_default(cls, pbar_cls: Callable | str):
        """
        Set the default progressbar class.

        This class method is useful when there is an user-defined class that
        follows ``_SupportProgress`` protocol.

        Parameters
        ----------
        pbar_cls : callable or str
            The default class. In principle this parameter does not have to be a
            class. As long as ``pbar_cls(max=...)`` returns a ``_SupportProgress``
            object it works. Either "default" or "napari" is also accepted.

        """
        if isinstance(pbar_cls, str):
            if pbar_cls == "napari":
                pbar_cls = NapariProgressBar
            elif pbar_cls == "default":
                pbar_cls = DefaultProgressBar
            else:
                raise ValueError(
                    f"Unknown progress bar {pbar_cls!r}. Must be either 'default' or "
                    "'napari', or a proper type object."
                )
        cls._DEFAULT_PROGRESS_BAR = pbar_cls
        return pbar_cls

    @staticmethod
    def to_callback(callback: Callable, *args, **kwargs) -> Callback:
        """Convert a callback to a callback object."""
        cb = Callback(callback)
        if args or kwargs:
            cb = cb(*args, **kwargs)
        return cb

    @property
    def is_generator(self) -> bool:
        """True if bound function is a generator function."""
        return inspect.isgeneratorfunction(self._func)

    @property
    def __doc__(self) -> str:
        """Synchronize docstring with bound function."""
        return self.func.__doc__

    @__doc__.setter
    def __doc__(self, doc: str):
        self.func.__doc__ = doc

    def _set_recorder(self, recorder: Callable[_P, Any]):
        """
        Set macro recorder function.
        Must accept ``recorder(bgui, *args, **kwargs)``.
        """
        self._recorder = recorder
        return None

    @overload
    def __call__(self, f: Callable[_P, _R1]) -> thread_worker:
        ...

    @overload
    def __call__(self, bgui: BaseGui, *args, **kwargs) -> Any:
        ...

    def __call__(self, *args, **kwargs):
        if self._func is None:
            f = args[0]
            self._func = f
            wraps(f)(self)  # NOTE: __name__ etc. are updated here.
            return self
        else:
            return self._func(*args, **kwargs)

    def __get__(self, gui: BaseGui, objtype=None):
        if gui is None:
            return self

        gui_id = id(gui)
        if gui_id in self._objects:
            return self._objects[gui_id]

        _create_worker = self._create_method(gui)
        _create_worker.__signature__ = self._get_method_signature()

        self._objects[gui_id] = _create_worker  # cache
        return _create_worker

    def _create_qt_worker(
        self, gui, *args, **kwargs
    ) -> FunctionWorker | GeneratorWorker:
        """Create a worker object."""
        worker = create_worker(
            self._func.__get__(gui),
            _ignore_errors=self._ignore_errors,
            _start_thread=False,
            *args,
            **kwargs,
        )
        return worker

    def _create_method(self, gui: BaseGui):
        from ..fields import MagicField

        @wraps(self)
        def _create_worker(*args, **kwargs):
            # create a worker object
            worker = self._create_qt_worker(gui, *args, **kwargs)
            is_generator = isinstance(worker, GeneratorWorker)

            if self._progress:
                _desc = self._progress["desc"]
                _total = self._progress["total"]
                _pbar = self._progress["pbar"]

                all_args = None
                if callable(_desc):
                    arguments = self.__signature__.bind(gui, *args, **kwargs)
                    arguments.apply_defaults()
                    all_args = arguments.arguments
                    desc = _desc(**all_args)
                else:
                    desc = str(_desc or self._func.__name__)

                if isinstance(_total, str):
                    if all_args is None:
                        arguments = self.__signature__.bind(gui, *args, **kwargs)
                        arguments.apply_defaults()
                        all_args = arguments.arguments
                    total = eval(_total, {}, all_args)
                elif callable(_total):
                    total = _total(gui)
                elif isinstance(_total, int):
                    total = _total
                else:
                    raise TypeError(
                        "'total' must be int, callable or evaluatable string."
                    )

                if not is_generator:
                    total = self._DEFAULT_TOTAL

                # create progressbar widget (or any proper widget)
                if _pbar is None:
                    pbar = self._find_progressbar(
                        gui,
                        desc=desc,
                        total=total,
                    )
                elif isinstance(_pbar, MagicField):
                    pbar = _pbar.get_widget(gui)
                    if not isinstance(pbar, ProgressBar):
                        raise TypeError(f"{_pbar.name} does not create a ProgressBar.")
                    pbar.label = desc or _pbar.name
                    pbar.max = total
                else:
                    pbar = _pbar
                    pbar.set_description(desc)

                worker.started.connect(init_pbar.__get__(pbar))

            self._bind_callbacks(worker, gui)

            # bind macro-recorder if exists
            if self._recorder is not None:
                worker.returned.connect(lambda _: self._recorder(gui, *args, **kwargs))

            if self._progress:
                if not getattr(pbar, "visible", False):
                    # return the progressbar to the initial state
                    worker.finished.connect(close_pbar.__get__(pbar))
                if pbar.max != 0 and is_generator:
                    worker.pbar = pbar  # avoid garbage collection
                    worker.yielded.connect(increment.__get__(pbar))

                if hasattr(pbar, "set_worker"):
                    # if _SupportProgress object support set_worker
                    pbar.set_worker(worker)

            _obj: PushButtonPlus | Action = gui[self._func.__name__]
            if _obj.running:
                worker.errored.connect(
                    partial(gui._error_mode.get_handler(), parent=gui)
                )
                if is_generator:
                    worker.aborted.connect(
                        gui._error_mode.wrap_handler(Aborted.raise_, parent=gui)
                    )

                worker.start()
            else:
                # If function is called from script, some events must get processed by
                # the application while keep script stopping at each line of code.
                app = use_app()
                worker.returned.connect(app.process_events)
                worker.started.connect(app.process_events)
                if isinstance(worker, GeneratorWorker):
                    worker.yielded.connect(app.process_events)
                worker.run()

            return None

        return _create_worker

    def _get_method_signature(self) -> inspect.Signature:
        sig = self.__signature__
        params = list(sig.parameters.values())[1:]
        return sig.replace(parameters=params)

    def _bind_callbacks(self, worker: FunctionWorker | GeneratorWorker, gui: BaseGui):
        # bind callbacks
        is_generator = isinstance(worker, GeneratorWorker)
        for c in self.started._iter_as_method(gui):
            worker.started.connect(c)
        for c in self.returned._iter_as_method(gui):
            worker.returned.connect(c)
        worker.returned.connect(partial(Callback.catch, macro=gui.macro))
        for c in self.errored._iter_as_method(gui):
            worker.errored.connect(c)
        for c in self.finished._iter_as_method(gui):
            worker.finished.connect(c)

        if is_generator:
            for c in self.aborted._iter_as_method(gui):
                worker.aborted.connect(c)
            for c in self.yielded._iter_as_method(gui):
                worker.yielded.connect(c)
            worker.yielded.connect(partial(Callback.catch, macro=gui.macro))

    @property
    def __signature__(self) -> inspect.Signature:
        """Get the signature of the bound function."""
        return get_signature(self._func)

    @__signature__.setter
    def __signature__(self, sig: inspect.Signature) -> None:
        """Update signature of the bound function."""
        if not isinstance(sig, inspect.Signature):
            raise TypeError(f"Cannot set type {type(sig)}.")
        self._func.__signature__ = sig
        return None

    def _find_progressbar(self, gui: BaseGui, desc: str | None = None, total: int = 0):
        """Find available progressbar. Create a new one if not found."""
        from ..fields import MagicField

        gui_id = id(gui)
        if gui_id in self._progressbars:
            _pbar = self._progressbars[gui_id]
        else:
            for name, attr in gui.__class__.__dict__.items():
                if isinstance(attr, MagicField):
                    attr = attr.get_widget(gui)
                if isinstance(attr, ProgressBar):
                    _pbar = self._progressbars[gui_id] = attr
                    if desc is None:
                        desc = name
                    break
            else:
                _pbar = self._progressbars[gui_id] = None
                if desc is None:
                    desc = "Progress"

        if _pbar is None:
            _pbar = self.__class__._DEFAULT_PROGRESS_BAR(max=total)
            if isinstance(_pbar, Widget) and _pbar.parent is None:
                # Popup progressbar as a splashscreen if it is not a child widget.
                _pbar.native.setParent(gui.native, self.__class__._WINDOW_FLAG)
                move_to_screen_center(_pbar.native)
        else:
            _pbar.max = total

        # try to set description
        if hasattr(_pbar, "set_description"):
            _pbar.set_description(desc)
        else:
            _pbar.label = desc
        return _pbar

    @property
    def started(self) -> CallbackList[None]:
        """Event that will be emitted on started."""
        return self._callback_dict_["started"]

    @property
    def returned(self) -> CallbackList[_R1]:
        """Event that will be emitted on returned."""
        return self._callback_dict_["returned"]

    @property
    def errored(self) -> CallbackList[Exception]:
        """Event that will be emitted on errored."""
        return self._callback_dict_["errored"]

    @property
    def yielded(self) -> CallbackList[_R1]:
        """Event that will be emitted on yielded."""
        if not self.is_generator:
            raise TypeError(
                f"Worker of non-generator function {self._func!r} does not have "
                "yielded signal."
            )
        return self._callback_dict_["yielded"]

    @property
    def finished(self) -> CallbackList[None]:
        """Event that will be emitted on finished."""
        return self._callback_dict_["finished"]

    @property
    def aborted(self) -> CallbackList[None]:
        """Event that will be emitted on aborted."""
        if not self.is_generator:
            raise TypeError(
                f"Worker of non-generator function {self._func!r} does not have "
                "aborted signal."
            )
        return self._callback_dict_["aborted"]


def init_pbar(pbar: ProgressBarLike):
    """Initialize progressbar."""
    pbar.value = 0
    pbar.show()
    return None


def close_pbar(pbar: ProgressBarLike):
    """Close progressbar."""
    if isinstance(pbar, ProgressBar):
        _labeled_widget = pbar._labeled_widget()
        if _labeled_widget is not None:
            pbar = _labeled_widget
    pbar.close()
    return None


def increment(pbar: ProgressBarLike):
    """Increment progressbar."""
    if pbar.value == pbar.max:
        pbar.max = 0
    else:
        pbar.value += 1
    return None


class Callback:
    """Callback object that can be recognized by thread_worker."""

    def __init__(self, f: Callable[[], Any]):
        if not callable(f):
            raise TypeError(f"{f} is not callable.")
        self._func = f

    @staticmethod
    def catch(out, macro: GuiMacro):
        if isinstance(out, Callback):
            with macro.blocked():
                out._func()

    def __call__(self, *args, **kwargs) -> Callback:
        """Return a partial callback."""
        return self.__class__(partial(self._func, *args, *kwargs))

    def __get__(self, obj, type=None) -> Callback:
        if obj is None:
            return self
        return self(obj)



from __future__ import annotations
from magicgui.widgets import Widget
from qtconsole.rich_jupyter_widget import RichJupyterWidget

# See napari_console
# https://github.com/napari/napari-console
class _Console(RichJupyterWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def connect_parent(self, ui: Widget):
        from IPython import get_ipython
        from IPython.terminal.interactiveshell import TerminalInteractiveShell
        from ipykernel.connect import get_connection_file
        from ipykernel.inprocess.ipkernel import InProcessInteractiveShell
        from ipykernel.zmqshell import ZMQInteractiveShell
        from qtconsole.client import QtKernelClient
        from qtconsole.inprocess import QtInProcessKernelManager

        if not isinstance(ui, Widget):
            raise TypeError(f"Cannot connect QtConsole to {type(ui)}.")

        shell = get_ipython()

        if shell is None:
            # If there is no currently running instance create an in-process
            # kernel.
            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel(show_banner=False)
            kernel_manager.kernel.gui = "qt"

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif type(shell) == InProcessInteractiveShell:
            # If there is an existing running InProcessInteractiveShell
            # it is likely because multiple viewers have been launched from
            # the same process. In that case create a new kernel.
            # Connect existing kernel
            kernel_manager = QtInProcessKernelManager(kernel=shell.kernel)
            kernel_client = kernel_manager.client()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif isinstance(shell, TerminalInteractiveShell):
            # if launching from an ipython terminal then adding a console is
            # not supported. Instead users should use the ipython terminal for
            # the same functionality.
            self.kernel_client = None
            self.kernel_manager = None
            self.shell = None
            self.push = lambda var: None

        elif isinstance(shell, ZMQInteractiveShell):
            # if launching from jupyter notebook, connect to the existing
            # kernel
            kernel_client = QtKernelClient(connection_file=get_connection_file())
            kernel_client.load_connection_file()
            kernel_client.start_channels()

            self.kernel_manager = None
            self.kernel_client = kernel_client
            self.shell = shell
            self.push = self.shell.push
        else:
            raise ValueError("ipython shell not recognized; " f"got {type(shell)}")

        self.parent_ui = ui
        self.shell.push({"ui": ui})
